from __future__ import annotations

import dataclasses
from pathlib import Path
from typing import Any, Callable, Mapping, Sequence, Tuple, cast

from ._format import prepare_progress, print_detail
from .runner_protocol import (
    BuildItemMapsFromAxisValuesFn,
    BuildItemMapsFromChunkOutputFn,
    CacheProtocol,
    CacheStatus,
    ChunkHashFn,
    CollectChunkDataFn,
    ExtractItemsFromMapFn,
    ItemHashFn,
    LoadChunkIndexFn,
    LoadPayloadFn,
    ReconstructOutputFromItemsFn,
    ReconstructPartialOutputFromItemsFn,
    ResolveCachePathFn,
    RunnerContext,
    UpdateChunkIndexFn,
    WriteChunkPayloadFn,
    WriteMetadataFn,
)

ChunkKey = Tuple[Tuple[str, Tuple[Any, ...]], ...]


@dataclasses.dataclass
class Diagnostics:
    """Execution diagnostics for run and parallel runners."""

    total_chunks: int = 0
    cached_chunks: int = 0
    executed_chunks: int = 0
    partial_chunks: int = 0
    merges: int = 0
    max_stream_items: int = 0
    stream_flushes: int = 0
    max_parallel_items: int = 0


@dataclasses.dataclass
class RunnerDeps:
    write_metadata: WriteMetadataFn | None = None
    chunk_hash: ChunkHashFn | None = None
    resolve_cache_path: ResolveCachePathFn | None = None
    load_payload: LoadPayloadFn | None = None
    write_chunk_payload: WriteChunkPayloadFn | None = None
    update_chunk_index: UpdateChunkIndexFn | None = None
    build_item_maps_from_axis_values: BuildItemMapsFromAxisValuesFn | None = None
    build_item_maps_from_chunk_output: BuildItemMapsFromChunkOutputFn | None = None
    reconstruct_output_from_items: ReconstructOutputFromItemsFn | None = None
    reconstruct_partial_output_from_items: (
        ReconstructPartialOutputFromItemsFn | None
    ) = None
    collect_chunk_data: CollectChunkDataFn | None = None
    item_hash: ItemHashFn | None = None
    load_chunk_index: LoadChunkIndexFn | None = None
    extract_items_from_map: ExtractItemsFromMapFn | None = None
    context: RunnerContext | None = None


def resolve_runner_deps(
    *,
    cache: CacheProtocol | None,
    context: RunnerContext | None = None,
    write_metadata: WriteMetadataFn | None = None,
    chunk_hash: ChunkHashFn | None = None,
    resolve_cache_path: ResolveCachePathFn | None = None,
    load_payload: LoadPayloadFn | None = None,
    write_chunk_payload: WriteChunkPayloadFn | None = None,
    update_chunk_index: UpdateChunkIndexFn | None = None,
    build_item_maps_from_axis_values: BuildItemMapsFromAxisValuesFn | None = None,
    build_item_maps_from_chunk_output: BuildItemMapsFromChunkOutputFn | None = None,
    reconstruct_output_from_items: ReconstructOutputFromItemsFn | None = None,
    reconstruct_partial_output_from_items: (
        ReconstructPartialOutputFromItemsFn | None
    ) = None,
    collect_chunk_data: CollectChunkDataFn | None = None,
    item_hash: ItemHashFn | None = None,
    load_chunk_index: LoadChunkIndexFn | None = None,
    extract_items_from_map: ExtractItemsFromMapFn | None = None,
    require: Sequence[str] | None = None,
) -> RunnerDeps:
    missing: list[str] = []
    require_set = set(require or ())

    def resolve_dep(name: str, value: Any, *, required: bool) -> Any:
        if value is not None:
            return value
        if cache is not None and hasattr(cache, name):
            return getattr(cache, name)
        if required:
            missing.append(name)
        return None

    context_resolved = context if context is not None else cache
    if context_resolved is None and "context" in require_set:
        missing.append("context")

    deps = RunnerDeps(
        write_metadata=resolve_dep(
            "write_metadata", write_metadata, required="write_metadata" in require_set
        ),
        chunk_hash=resolve_dep(
            "chunk_hash", chunk_hash, required="chunk_hash" in require_set
        ),
        resolve_cache_path=resolve_dep(
            "resolve_cache_path",
            resolve_cache_path,
            required="resolve_cache_path" in require_set,
        ),
        load_payload=resolve_dep(
            "load_payload", load_payload, required="load_payload" in require_set
        ),
        write_chunk_payload=resolve_dep(
            "write_chunk_payload",
            write_chunk_payload,
            required="write_chunk_payload" in require_set,
        ),
        update_chunk_index=resolve_dep(
            "update_chunk_index",
            update_chunk_index,
            required="update_chunk_index" in require_set,
        ),
        build_item_maps_from_axis_values=resolve_dep(
            "build_item_maps_from_axis_values",
            build_item_maps_from_axis_values,
            required="build_item_maps_from_axis_values" in require_set,
        ),
        build_item_maps_from_chunk_output=resolve_dep(
            "build_item_maps_from_chunk_output",
            build_item_maps_from_chunk_output,
            required="build_item_maps_from_chunk_output" in require_set,
        ),
        reconstruct_output_from_items=resolve_dep(
            "reconstruct_output_from_items",
            reconstruct_output_from_items,
            required="reconstruct_output_from_items" in require_set,
        ),
        reconstruct_partial_output_from_items=resolve_dep(
            "reconstruct_partial_output_from_items",
            reconstruct_partial_output_from_items,
            required="reconstruct_partial_output_from_items" in require_set,
        ),
        collect_chunk_data=resolve_dep(
            "collect_chunk_data",
            collect_chunk_data,
            required="collect_chunk_data" in require_set,
        ),
        item_hash=resolve_dep(
            "item_hash", item_hash, required="item_hash" in require_set
        ),
        load_chunk_index=resolve_dep(
            "load_chunk_index",
            load_chunk_index,
            required="load_chunk_index" in require_set,
        ),
        extract_items_from_map=resolve_dep(
            "extract_items_from_map",
            extract_items_from_map,
            required="extract_items_from_map" in require_set,
        ),
        context=context_resolved,
    )

    if missing:
        missing = list(dict.fromkeys(missing))
        raise ValueError(
            "Missing runner dependencies: "
            + ", ".join(missing)
            + ". Provide them explicitly or pass cache=..."
        )

    return deps


def resolve_cache_for_run(
    cache: Any,
    *,
    params: dict[str, Any] | None = None,
    axis_values_override: Mapping[str, Sequence[Any]] | None = None,
    extend_cache: bool = False,
    allow_superset: bool = False,
) -> CacheProtocol:
    if not hasattr(cache, "cache_for_params"):
        return cast(CacheProtocol, cache)
    if params is None:
        raise ValueError("params is required when using ChunkMemo with runners")
    axis_values = dict(axis_values_override) if axis_values_override else None
    axis_values_for_load = None
    allow_superset_for_load = allow_superset
    if extend_cache:
        axis_values_for_load = cache.axis_values
        allow_superset_for_load = True
    elif axis_values is not None and allow_superset:
        axis_values_for_load = axis_values
    if axis_values_for_load is not None:
        cache = cache.auto_load(
            root=cache.root,
            params=params,
            axis_values=axis_values_for_load,
            chunk_spec=cache.chunk_spec,
            metadata=cache.metadata,
            collate_fn=cache.collate_fn,
            chunk_enumerator=cache.chunk_enumerator,
            chunk_hash_fn=cache.chunk_hash_fn,
            path_fn=cache.path_fn,
            version=cache.version,
            axis_order=cache.axis_order,
            verbose=cache.verbose,
            profile=cache.profile,
            exclusive=cache.exclusive,
            warn_on_overlap=cache.warn_on_overlap,
            precompute_chunk_keys=cache.precompute_chunk_keys,
            allow_superset=allow_superset_for_load,
        )
    cache = cache.cache_for_params(params)
    if extend_cache and axis_values is not None:
        cache.cache_status(axis_values_override=axis_values, extend_cache=True)
    return cast(CacheProtocol, cache)


def _stream_item_count(output: Any) -> int:
    if isinstance(output, (list, tuple, dict)):
        return len(output)
    return 1


def _format_item_count(count: int | None) -> str:
    return "all" if count is None else str(count)


def _log_chunk(
    context: RunnerContext, action: str, chunk_key: ChunkKey, item_count: int | None
) -> None:
    if context.verbose >= 2:
        print_detail(
            f"[ChunkMemo] {action} chunk={chunk_key} items={_format_item_count(item_count)}"
        )


def _payload_item_map(
    *,
    build_item_maps_from_chunk_output: BuildItemMapsFromChunkOutputFn,
    resolve_cache_path: ResolveCachePathFn,
    write_chunk_payload: WriteChunkPayloadFn,
    chunk_key: ChunkKey,
    payload: dict[str, Any],
    chunk_hash: str | None = None,
    write_back: bool = False,
) -> dict[str, Any] | None:
    item_map = payload.get("items")
    if item_map is None:
        item_map, item_axis_vals = build_item_maps_from_chunk_output(
            chunk_key,
            chunk_output=payload.get("output"),
        )
        if item_map is not None:
            payload["items"] = item_map
            if item_axis_vals is not None:
                payload["axis_vals"] = item_axis_vals
            if write_back:
                if chunk_hash is None:
                    raise ValueError("chunk_hash required for write_back")
                path = resolve_cache_path(chunk_key, chunk_hash)
                write_chunk_payload(path, payload, existing=payload)
    return item_map


def resolve_chunk_path(
    chunk_hash: ChunkHashFn,
    resolve_cache_path: ResolveCachePathFn,
    chunk_key: ChunkKey,
) -> tuple[str, Path]:
    chunk_hash_value = chunk_hash(chunk_key)
    path = resolve_cache_path(chunk_key, chunk_hash_value)
    return chunk_hash_value, path


def _save_chunk_payload(
    *,
    resolve_cache_path: ResolveCachePathFn,
    write_chunk_payload: WriteChunkPayloadFn,
    update_chunk_index: UpdateChunkIndexFn,
    chunk_key: ChunkKey,
    chunk_output: Any,
    item_map: dict[str, Any] | None,
    cached_payloads: dict[ChunkKey, Mapping[str, Any]],
    chunk_hash: str,
    missing_chunks: list[ChunkKey],
    spec_fn: Callable[..., dict[str, Any]] | None = None,
) -> dict[str, Any]:
    if chunk_key in missing_chunks:
        payload: dict[str, Any] = {}
        if item_map is not None:
            payload["items"] = item_map
        else:
            payload["output"] = chunk_output
    else:
        existing = cached_payloads.get(chunk_key, {})
        payload = dict(existing)
        existing_items = existing.get("items", {})
        merged_items = dict(existing_items)
        if item_map is not None:
            merged_items.update(item_map)
        payload["items"] = merged_items
        if "output" not in payload and chunk_output is not None and item_map is None:
            payload["output"] = chunk_output
    if spec_fn is not None:
        payload["axis_vals"] = spec_fn()
    path = resolve_cache_path(chunk_key, chunk_hash)
    write_chunk_payload(
        path,
        payload,
        existing=cached_payloads.get(chunk_key),
    )
    update_chunk_index(chunk_hash, chunk_key)
    return payload


def _require_params(cache_status: CacheStatus) -> dict[str, Any]:
    metadata = cache_status.get("metadata")
    if isinstance(metadata, Mapping):
        params = metadata.get("params")
        if isinstance(params, Mapping):
            return dict(params)
    return {}


def _require_axis_info(
    cache_status: CacheStatus,
) -> tuple[Tuple[str, ...], dict[str, dict[Any, int]]]:
    axis_order = cache_status.get("axis_order")
    if not isinstance(axis_order, (tuple, list)):
        raise ValueError("cache_status must include axis_order")
    axis_chunk_maps = cache_status.get("axis_chunk_maps")
    if not isinstance(axis_chunk_maps, Mapping):
        raise ValueError("cache_status must include axis_chunk_maps")
    return tuple(axis_order), cast(dict[str, dict[Any, int]], axis_chunk_maps)


def _merge_outputs(
    context: RunnerContext,
    outputs: list[Any],
    diagnostics: Diagnostics,
    collate_fn: Callable[[list[Any]], Any] | None = None,
) -> Any:
    diagnostics.merges += 1
    if collate_fn is not None:
        return collate_fn(outputs)
    if context.collate_fn is not None:
        return context.collate_fn(outputs)
    return outputs


def prepare_planning_progress(
    *,
    total_chunks: int,
    total_items: int,
    verbose: int,
    label: str = "planning",
) -> tuple[Callable[[int, bool], None], Callable[[int], None], Callable[[bool], None]]:
    """Prepare planning progress callbacks with gated final output."""
    report_base, update_processed = prepare_progress(
        total_chunks=total_chunks,
        total_items=total_items,
        verbose=verbose,
        label=label,
    )
    allow_final = False
    final_emitted = False

    def set_allow_final(value: bool) -> None:
        nonlocal allow_final
        allow_final = value

    def report_progress(processed: int, final: bool = False) -> None:
        nonlocal final_emitted
        if processed >= total_chunks:
            if not allow_final or final_emitted:
                return
            final_emitted = True
            report_base(total_chunks, True)
            return
        if final:
            return
        report_base(processed, False)

    return report_progress, update_processed, set_allow_final



import hashlib
import functools
import inspect
import itertools
import json
import pickle
import time
import warnings
from collections.abc import Sequence as ABCSequence
from pathlib import Path
from typing import Any, Callable, Mapping, Sequence, Tuple, cast

from .cache_index import load_chunk_index, update_chunk_index
from .data_write_utils import (
    _apply_payload_timestamps,
    _atomic_write_json,
    _atomic_write_pickle,
)
from .identity import stable_serialize
from .runner_protocol import CacheStatus

ChunkKey = Tuple[Tuple[str, Tuple[Any, ...]], ...]
MemoChunkEnumerator = Callable[[dict[str, Any]], Sequence[ChunkKey]]
CollateFn = Callable[[list[Any]], Any]
AxisIndexMap = dict[str, dict[Any, int]]
CachePathFn = Callable[[dict[str, Any], ChunkKey, str, str], Path | str]


def default_chunk_hash(cache_id: str, chunk_key: ChunkKey, version: str) -> str:
    payload = {
        "cache_id": cache_id,
        "chunk_key": chunk_key,
        "version": version,
    }
    data = stable_serialize(payload)
    return hashlib.sha256(data.encode("utf-8")).hexdigest()


def _chunk_values(values: Sequence[Any], size: int) -> list[Tuple[Any, ...]]:
    if size <= 0:
        raise ValueError("chunk size must be > 0")
    chunks: list[Tuple[Any, ...]] = []
    for start in range(0, len(values), size):
        end = min(start + size, len(values))
        chunks.append(tuple(values[start:end]))
    return chunks


class ChunkCache:
    def __init__(
        self,
        root: str | Path | None,
        chunk_spec: dict[str, Any] | None,
        axis_values: dict[str, Any],
        cache_id: str,
        metadata: dict[str, Any] | None = None,
        collate_fn: CollateFn | None = None,
        chunk_enumerator: MemoChunkEnumerator | None = None,
        chunk_hash_fn: Callable[[str, ChunkKey, str], str] | None = None,
        path_fn: CachePathFn | None = None,
        version: str = "v1",
        axis_order: Sequence[str] | None = None,
        verbose: int = 1,
        profile: bool = False,
        exclusive: bool = False,
        warn_on_overlap: bool = False,
        precompute_chunk_keys: bool = False,
    ) -> None:
        """Initialize a ChunkCache.

        Args:
            root: Directory for chunk cache files.
            chunk_spec: Per-axis chunk sizes (e.g., {"strat": 1, "s": 3}).
                If None, defaults to size 1 for each axis.
            collate_fn: Optional collate function for the list of chunk outputs.
            chunk_enumerator: Optional chunk enumerator that defines the
                cache chunk order.
            chunk_hash_fn: Optional override for chunk hashing.
            path_fn: Optional override for cache file paths. This hook is
                experimental and not yet thoroughly tested.
            version: Cache namespace/version tag.
            axis_order: Axis iteration order (defaults to lexicographic).
            axis_values: Canonical axis values for the cache. Can be:
                - dict of iterables (lists, tuples): converted to list on init
                - dict of callables: For memory-efficient lazy loading.
                  Supports two patterns:
                  * Index-based: `lambda idx: values[idx]` - called with index
                  * List-returning: `lambda: full_list` - called once for all values
                  Use callables for large datasets to avoid loading all values into memory.
            verbose: Verbosity flag.
            profile: Enable profiling output.
            cache_id: Cache identity string (typically a hash of params).
            metadata: Optional metadata dict to store alongside the cache.
            exclusive: If True, error when creating a cache that conflicts
                with existing cache metadata for this cache_id.
            warn_on_overlap: If True, warn when caches overlap for this cache_id.

        Example:
            >>> cache = ChunkCache(
            ...     root="/tmp/chunk-memo",
            ...     chunk_spec={"strat": 1, "s": 3},
            ...     axis_values={"strat": ["a"], "s": [1, 2, 3]},
            ... )
            >>> cache = ChunkCache(
            ...     root="/tmp/chunk-memo",
            ...     cache_id="example",
            ...     chunk_spec={"strat": 1, "s": 3},
            ...     axis_values={"strat": ["a"], "s": [1, 2, 3]},
            ... )
            >>> for chunk_key, chunk_hash in cache.chunk_hashes_for():
            ...     path = cache.resolve_cache_path(chunk_key, chunk_hash)
            ...     payload = {"output": []}
            ...     cache.write_chunk_payload(path, payload)
        """
        root_path = Path(root) if root is not None else None
        self.cache_id = cache_id
        self.metadata = metadata or {}
        if chunk_spec is None:
            chunk_spec = {
                axis: (
                    len(axis_values[axis])
                    if isinstance(axis_values[axis], (list, tuple))
                    else 1
                )
                for axis in axis_values
            }
        self.chunk_spec = chunk_spec
        self.collate_fn = collate_fn
        self.chunk_enumerator = chunk_enumerator
        self.chunk_hash_fn = chunk_hash_fn or default_chunk_hash
        self.path_fn = path_fn
        self.version = version
        self.axis_order = tuple(axis_order) if axis_order is not None else None
        self.verbose = verbose
        self.profile = profile
        self.exclusive = exclusive
        self.warn_on_overlap = warn_on_overlap
        self.precompute_chunk_keys = precompute_chunk_keys
        self._axis_values: dict[str, list[Any]] | None = None
        self._axis_index_map: AxisIndexMap | None = None
        self._axis_values_serializable: dict[str, Any] | None = None
        self._normalized_chunk_keys: list[ChunkKey] | None = None
        self._selected_items_by_chunk: (
            Mapping[ChunkKey, list[Tuple[Any, ...]]] | None
        ) = None
        self._checked_exclusive = False
        self._set_axis_values(axis_values)
        self._axis_values_serializable = self._make_axis_values_serializable()
        if root_path is None:
            root_path = Path.cwd() / ".chunk_memo"
        if not self.cache_id:
            raise ValueError("cache_id must be a non-empty string")
        self._memo_root_override: Path | None = None
        metadata_path = root_path / "metadata.json"
        if metadata_path.exists():
            self._memo_root_override = root_path
            self.root = root_path.parent
        else:
            self.root = root_path
        self._memo_root().mkdir(parents=True, exist_ok=True)
        if self.exclusive or self.warn_on_overlap:
            self._check_exclusive()
        if self.precompute_chunk_keys:
            self._normalized_chunk_keys = self._build_chunk_keys()

    def cache_status(
        self,
        *,
        axis_indices: Mapping[str, Any] | None = None,
        axis_values_override: Mapping[str, Sequence[Any]] | None = None,
        extend_cache: bool = False,
        **axes: Any,
    ) -> CacheStatus:
        """Return cached vs missing chunk info for a subset of axes.

        axis_indices selects axes by index (int, slice, range, or list/tuple of
        those), based on the canonical axis order. Chunk indices are
        returned as lists of ints.
        axis_values_override provides explicit axis values for axis_indices
        expansion; when extend_cache=True those values are appended in-place.
        """
        profile_start = time.monotonic() if self.profile else None
        if self._axis_values is None:
            raise ValueError("axis_values must be set before checking cache status")
        if axis_indices is not None and axes:
            raise ValueError("axis_indices cannot be combined with axis values")
        if axis_values_override is not None and extend_cache:
            self.extend_axis_values(axis_values_override)
        axis_values = self._normalize_axes(
            axes,
            axis_indices=axis_indices,
            axis_values_override=axis_values_override,
        )
        axis_order = self._resolve_axis_order(dict(axis_values))
        axis_chunk_maps = self._axis_chunk_maps(axis_values, axis_order)
        if axis_indices is None and not axes:
            chunk_keys = self._build_chunk_keys(axis_values)
        else:
            chunk_keys, _ = self._build_chunk_plan_for_axes(axis_values)
        chunk_index = self.load_chunk_index()
        use_index = bool(chunk_index)
        if self.profile and self.verbose >= 1 and profile_start is not None:
            print(
                f"[ChunkCache] profile cache_status_build_s={time.monotonic() - profile_start:0.3f}"
            )
        cached_chunks: list[ChunkKey] = []
        cached_chunk_indices: list[dict[str, Any]] = []
        missing_chunks: list[ChunkKey] = []
        missing_chunk_indices: list[dict[str, Any]] = []
        for chunk_key in chunk_keys:
            chunk_hash = self.chunk_hash(chunk_key)
            indices = self._chunk_indices_from_key(chunk_key)
            if use_index:
                exists = chunk_hash in chunk_index
            else:
                path = self.resolve_cache_path(chunk_key, chunk_hash)
                exists = path.exists()
            if exists:
                cached_chunks.append(chunk_key)
                cached_chunk_indices.append(indices)
            else:
                missing_chunks.append(chunk_key)
                missing_chunk_indices.append(indices)
        if self.profile and self.verbose >= 1 and profile_start is not None:
            print(
                f"[ChunkCache] profile cache_status_scan_s={time.monotonic() - profile_start:0.3f}"
            )
        payload = {
            "cache_id": self.cache_id,
            "metadata": self.metadata,
            "axis_values": axis_values,
            "axis_order": axis_order,
            "axis_chunk_maps": axis_chunk_maps,
            "total_chunks": len(chunk_keys),
            "cached_chunks": cached_chunks,
            "cached_chunk_indices": cached_chunk_indices,
            "missing_chunks": missing_chunks,
            "missing_chunk_indices": missing_chunk_indices,
        }
        return cast(CacheStatus, payload)

    def slice(
        self,
        axis_indices: Mapping[str, Any] | None = None,
        **axes: Any,
    ) -> "ChunkCache":
        """Return a sub-cache scoped to the requested axes."""
        if self._axis_values is None:
            raise ValueError("axis_values must be set before slicing")
        if not self._axis_values and (axes or axis_indices):
            raise ValueError(
                "Cannot pass axis arguments to a singleton cache (no axes)"
            )
        if axis_indices is not None and axes:
            raise ValueError("axis_indices cannot be combined with axis values")
        selected_items: Mapping[ChunkKey, list[Tuple[Any, ...]]] | None = None
        normalized_chunk_keys: list[ChunkKey] | None = None
        if axis_indices is None and not axes:
            normalized_axis_values = self._axis_values
            if self.precompute_chunk_keys:
                normalized_chunk_keys = self._build_chunk_keys()
        else:
            normalized_axis_values = self._normalize_axes(
                axes, axis_indices=axis_indices
            )
            normalized_chunk_keys, selected_items = self._build_chunk_plan_for_axes(
                normalized_axis_values
            )
        sliced = ChunkCache(
            root=self.root,
            chunk_spec=self.chunk_spec,
            axis_values=normalized_axis_values,
            cache_id=self.cache_id,
            metadata=self.metadata,
            collate_fn=self.collate_fn,
            chunk_enumerator=self.chunk_enumerator,
            chunk_hash_fn=self.chunk_hash_fn,
            path_fn=self.path_fn,
            version=self.version,
            axis_order=self.axis_order,
            verbose=self.verbose,
            profile=self.profile,
            exclusive=self.exclusive,
            warn_on_overlap=self.warn_on_overlap,
            precompute_chunk_keys=self.precompute_chunk_keys,
        )
        sliced._memo_root_override = self._memo_root()
        sliced._selected_items_by_chunk = selected_items
        if normalized_chunk_keys is not None:
            sliced._normalized_chunk_keys = normalized_chunk_keys
        if sliced.exclusive:
            sliced._check_exclusive()
        return sliced

    def bind_exec_fn(self, exec_fn: Callable[..., Any]) -> Callable[..., Any]:
        """Return exec_fn (no binding at cache level)."""
        return self._bind_exec_fn(exec_fn)

    def resolved_chunk_keys(self) -> list[ChunkKey]:
        """Return precomputed chunk keys if available, else build them."""
        if self._normalized_chunk_keys is not None:
            return list(self._normalized_chunk_keys)
        return self._build_chunk_keys()

    def requested_items_by_chunk(
        self,
    ) -> Mapping[ChunkKey, list[Tuple[Any, ...]]] | None:
        """Return requested items per chunk for a sliced cache, if any."""
        return self._selected_items_by_chunk

    def resolve_cache_path(self, chunk_key: ChunkKey, chunk_hash: str) -> Path:
        memo_root = self._memo_root()
        if self.path_fn is None:
            return memo_root / "chunks" / f"{chunk_hash}.pkl"
        path = self.path_fn(self.metadata, chunk_key, self.version, chunk_hash)
        path_obj = Path(path)
        if path_obj.is_absolute():
            return path_obj
        return memo_root / path_obj

    def _chunk_keys_for(
        self,
        *,
        axis_indices: Mapping[str, Any] | None = None,
        **axes: Any,
    ) -> list[ChunkKey]:
        if self._axis_values is None:
            raise ValueError("axis_values must be set before building chunk keys")
        if not self._axis_values and (axes or axis_indices):
            raise ValueError(
                "Cannot pass axis arguments to a singleton cache (no axes)"
            )
        if axis_indices is None and not axes:
            return self._build_chunk_keys()
        if axis_indices is not None and axes:
            raise ValueError("axis_indices cannot be combined with axis values")
        axis_values = self._normalize_axes(axes, axis_indices=axis_indices)
        chunk_keys, _ = self._build_chunk_plan_for_axes(axis_values)
        return chunk_keys

    def chunk_hashes_for(
        self,
        *,
        axis_indices: Mapping[str, Any] | None = None,
        **axes: Any,
    ) -> list[tuple[ChunkKey, str]]:
        """Return chunk keys with their hashes for the requested axes."""
        chunk_keys = self._chunk_keys_for(axis_indices=axis_indices, **axes)
        return [(chunk_key, self.chunk_hash(chunk_key)) for chunk_key in chunk_keys]

    def apply_payload_timestamps(
        self,
        payload: dict[str, Any],
        *,
        existing: Mapping[str, Any] | None = None,
    ) -> None:
        _apply_payload_timestamps(payload, existing=existing)

    def write_chunk_payload(
        self,
        path: Path,
        payload: dict[str, Any],
        *,
        existing: Mapping[str, Any] | None = None,
    ) -> Path:
        self.apply_payload_timestamps(payload, existing=existing)
        _atomic_write_pickle(path, payload)
        return path

    def write_json(
        self,
        path: Path,
        payload: dict[str, Any],
        *,
        existing: Mapping[str, Any] | None = None,
    ) -> Path:
        if existing is not None:
            self.apply_payload_timestamps(payload, existing=existing)
        _atomic_write_json(path, payload)
        return path

    def load_chunk_index(self) -> dict[str, Any]:
        return load_chunk_index(self._memo_root())

    def update_chunk_index(self, chunk_hash: str, chunk_key: ChunkKey) -> None:
        update_chunk_index(self._memo_root(), chunk_hash, chunk_key)

    def chunk_hash(self, chunk_key: ChunkKey) -> str:
        return self.chunk_hash_fn(self.cache_id, chunk_key, self.version)

    def cache_hash(self) -> str:
        return self.cache_id

    def _check_exclusive(self) -> None:
        if self._checked_exclusive:
            return
        self._checked_exclusive = True
        if not self.exclusive and not self.warn_on_overlap:
            return

        if self._axis_values is None:
            return
        if self._axis_values_serializable is None:
            return
        axis_values_serializable = self._axis_values_serializable

        metadata_path = self._memo_root() / "metadata.json"
        existing_meta = self._read_metadata_file(metadata_path, required=False)
        if existing_meta is None:
            return
        meta_axis_values = existing_meta.get("axis_values", {})
        meta_chunk_spec = existing_meta.get("chunk_spec")
        meta_axis_order = existing_meta.get("axis_order")
        meta_version = existing_meta.get("version")

        if self.exclusive:
            if meta_axis_values != axis_values_serializable:
                raise ValueError(
                    "Cache with same cache_id already exists with different axis_values. "
                    f"Existing axis_values={meta_axis_values}, requested={axis_values_serializable}. "
                    "Use a different cache_id, or set exclusive=False."
                )
            if meta_chunk_spec != self.chunk_spec:
                raise ValueError(
                    "Cache with same cache_id already exists with different chunk_spec. "
                    f"Existing chunk_spec={meta_chunk_spec}, requested={self.chunk_spec}. "
                    "Use a different cache_id, or set exclusive=False."
                )
            if meta_axis_order != self.axis_order:
                raise ValueError(
                    "Cache with same cache_id already exists with different axis_order. "
                    f"Existing axis_order={meta_axis_order}, requested={self.axis_order}. "
                    "Use a different cache_id, or set exclusive=False."
                )
            if meta_version is not None and meta_version != self.version:
                raise ValueError(
                    "Cache with same cache_id already exists with different version. "
                    f"Existing version={meta_version}, requested={self.version}. "
                    "Use a different cache_id, or set exclusive=False."
                )

        if self.warn_on_overlap and meta_axis_values and axis_values_serializable:
            overlap = self._detect_axis_overlap(
                meta_axis_values, axis_values_serializable
            )
            if overlap:
                warnings.warn(
                    "Cache overlap detected: "
                    f"existing axis_values={meta_axis_values}, "
                    f"current axis_values={axis_values_serializable}. Overlap: {overlap}. "
                    "Consider using exclusive=True to prevent accidental conflicts.",
                    stacklevel=2,
                )

    def _detect_axis_overlap(
        self, axis_values_a: dict[str, Any], axis_values_b: dict[str, Any]
    ) -> dict[str, list[Any]] | None:
        shared_axes = set(axis_values_a) & set(axis_values_b)
        if not shared_axes:
            return None
        overlap: dict[str, list[Any]] = {}
        for axis in shared_axes:
            values_a = self._axis_value_set(axis_values_a[axis])
            values_b = self._axis_value_set(axis_values_b[axis])
            intersection = values_a & values_b
            if intersection:
                overlap[axis] = sorted(intersection)
            else:
                return None
        return overlap

    def _memo_root(self) -> Path:
        if self._memo_root_override is not None:
            return self._memo_root_override
        return self.root / self.cache_id

    def write_metadata(self) -> Path:
        memo_root = self._memo_root()
        memo_root.mkdir(parents=True, exist_ok=True)
        path = memo_root / "metadata.json"
        existing_meta = None
        if path.exists():
            try:
                with open(path, "r", encoding="utf-8") as handle:
                    existing_meta = json.load(handle)
            except json.JSONDecodeError:
                existing_meta = None
        payload = {
            "cache_id": self.cache_id,
            "metadata": self.metadata,
            "axis_values": self._axis_values_serializable,
            "chunk_spec": self.chunk_spec,
            "version": self.version,
            "axis_order": self.axis_order,
        }
        self.write_json(path, payload, existing=existing_meta)
        return path

    def set_identity(
        self, cache_id: str, metadata: dict[str, Any] | None = None
    ) -> None:
        """Update cache identity and refresh memo root."""
        if not cache_id:
            raise ValueError("cache_id must be a non-empty string")
        self.cache_id = cache_id
        if metadata is not None:
            self.metadata = metadata
        self._checked_exclusive = False
        if self._memo_root_override is None:
            self._memo_root().mkdir(parents=True, exist_ok=True)
        if self.exclusive:
            self._check_exclusive()

    def _bind_exec_fn(self, exec_fn: Callable[..., Any]) -> Callable[..., Any]:
        return exec_fn

    def _resolve_axis_order(self, axis_values: dict[str, Any]) -> Tuple[str, ...]:
        if self.axis_order is not None:
            return self.axis_order
        return tuple(sorted(axis_values))

    def _ordered_axis_values(
        self,
        axis_values: Mapping[str, Sequence[Any]],
        axis_order: Sequence[str] | None = None,
    ) -> tuple[Tuple[str, ...], list[tuple[str, Sequence[Any], int]]]:
        if axis_order is None:
            axis_order = self._resolve_axis_order(dict(axis_values))
        ordered: list[tuple[str, Sequence[Any], int]] = []
        for axis in axis_order:
            values = axis_values.get(axis)
            if values is None:
                raise KeyError(f"Missing axis '{axis}' in axis_values")
            size = self._resolve_axis_chunk_size(axis)
            ordered.append((axis, values, size))
        return tuple(axis_order), ordered

    def _ordered_axis_requests(
        self,
        axis_values: Mapping[str, Sequence[Any]],
        axis_order: Sequence[str],
    ) -> list[tuple[str, list[Any], int]]:
        ordered: list[tuple[str, list[Any], int]] = []
        for axis in axis_order:
            requested_values = list(axis_values.get(axis, []))
            if not requested_values:
                raise ValueError(f"Missing values for axis '{axis}'")
            size = self._resolve_axis_chunk_size(axis)
            ordered.append((axis, requested_values, size))
        return ordered

    def _axis_chunk_maps(
        self, axis_values: Mapping[str, Any], axis_order: Sequence[str]
    ) -> dict[str, dict[Any, int]]:
        axis_chunk_maps: dict[str, dict[Any, int]] = {}
        _, ordered = self._ordered_axis_values(axis_values, axis_order)
        for axis, values, size in ordered:
            value_to_chunk_id: dict[Any, int] = {}
            for index, value in enumerate(values):
                value_to_chunk_id[value] = index // size
            axis_chunk_maps[axis] = value_to_chunk_id
        return axis_chunk_maps

    def _set_axis_values(self, axis_values: dict[str, Any]) -> None:
        axis_order = self._resolve_axis_order(axis_values)
        for axis in axis_order:
            if axis not in axis_values:
                raise KeyError(f"Missing axis '{axis}' in axis_values")

        axis_index_map: AxisIndexMap = {}
        axis_value_lists: dict[str, list[Any]] = {}
        for axis in axis_order:
            axis_values_obj = axis_values[axis]
            values = self._materialize_axis_values(axis_values_obj)
            axis_value_lists[axis] = values
            axis_index_map[axis] = {value: index for index, value in enumerate(values)}

        self._axis_values = axis_value_lists
        self._axis_index_map = axis_index_map

    def _make_axis_values_serializable(self) -> dict[str, Any]:
        """Convert axis_values to a JSON-serializable representation."""
        if self._axis_values is None:
            raise ValueError("axis_values must be set before serializing")
        axis_order = self._resolve_axis_order(self._axis_values)
        return {axis: list(self._axis_values[axis]) for axis in axis_order}

    def _get_all_axis_values(self, axis: str) -> list[Any]:
        """Get all values for an axis."""
        if self._axis_values is None:
            raise ValueError("axis_values must be set before getting axis values")
        return list(self._axis_values[axis])

    def extend_axis_values(
        self,
        axis_values: Mapping[str, Sequence[Any]],
        *,
        write_metadata: bool = True,
    ) -> None:
        """Extend axis_values in-place with new values.

        Values already present are ignored; new values are appended in order.
        When write_metadata=True, metadata.json is updated immediately.
        """
        if self._axis_values is None or self._axis_index_map is None:
            raise ValueError("axis_values must be set before extending")
        axis_order = self._resolve_axis_order(self._axis_values)
        missing_axes = [axis for axis in axis_values if axis not in self._axis_values]
        if missing_axes:
            raise KeyError(f"Unknown axis(es) in extension: {missing_axes}")

        updated = False
        for axis in axis_order:
            if axis not in axis_values:
                continue
            current_values = self._axis_values[axis]
            current_set = set(current_values)
            new_values: list[Any] = []
            for value in axis_values[axis]:
                if value in current_set:
                    continue
                current_set.add(value)
                new_values.append(value)
            if not new_values:
                continue
            updated = True
            current_values.extend(new_values)
            self._axis_index_map[axis] = {
                value: index for index, value in enumerate(current_values)
            }

        if updated:
            self._axis_values_serializable = self._make_axis_values_serializable()
            if write_metadata:
                self.write_metadata()

    def _normalize_axes(
        self,
        axes: Mapping[str, Any] | None,
        *,
        axis_indices: Mapping[str, Any] | None = None,
        axis_values_override: Mapping[str, Sequence[Any]] | None = None,
    ) -> dict[str, list[Any]]:
        if axis_indices is not None:
            if axes:
                raise ValueError("axis_indices cannot be combined with axis values")
            return self._normalize_axis_indices(
                axis_indices,
                axis_values_override=axis_values_override,
            )
        return self._normalize_axis_values(axes or {})

    def _normalize_axis_values(self, axes: Mapping[str, Any]) -> dict[str, list[Any]]:
        """Normalize axis value selections into full axis lists."""
        if self._axis_values is None or self._axis_index_map is None:
            raise ValueError("axis_values must be set before running memoized function")
        axis_values: dict[str, list[Any]] = {}
        for axis in self._axis_values:
            if axis not in axes:
                axis_values[axis] = self._get_all_axis_values(axis)
                continue
            values = axes[axis]
            if isinstance(values, (list, tuple)):
                normalized = list(values)
            else:
                normalized = [values]
            for value in normalized:
                if value not in self._axis_index_map[axis]:
                    raise KeyError(
                        f"Value '{value}' not found in axis_values for axis '{axis}'"
                    )
            axis_values[axis] = normalized
        return axis_values

    def _normalize_axis_indices(
        self,
        axis_indices: Mapping[str, Any],
        *,
        axis_values_override: Mapping[str, Sequence[Any]] | None = None,
    ) -> dict[str, list[Any]]:
        """Normalize axis_indices into axis values using canonical axis order."""
        if self._axis_values is None:
            raise ValueError("axis_values must be set before running memoized function")
        axis_values: dict[str, list[Any]] = {}
        for axis in self._axis_values:
            if axis not in axis_indices:
                axis_values[axis] = self._get_all_axis_values(axis)
                continue
            values = axis_indices[axis]
            if axis_values_override is not None and axis in axis_values_override:
                axis_values_obj = axis_values_override[axis]
            else:
                axis_values_obj = self._axis_values[axis]
            indices = self._expand_axis_indices(values, axis, len(axis_values_obj))
            axis_values[axis] = [axis_values_obj[index] for index in indices]
        return axis_values

    def _expand_axis_indices(self, values: Any, axis: str, axis_len: int) -> list[int]:
        if isinstance(values, (list, tuple)):
            items = list(values)
        else:
            items = [values]
        resolved: list[int] = []
        for item in items:
            if isinstance(item, range):
                indices = list(item)
            elif isinstance(item, slice):
                start, stop, step = item.indices(axis_len)
                indices = list(range(start, stop, step))
            elif isinstance(item, int):
                indices = [item]
            else:
                raise TypeError(
                    f"Indices for axis '{axis}' must be int, slice, or range, got {type(item)}"
                )
            for index in indices:
                if index < 0 or index >= axis_len:
                    raise IndexError(f"Index {index} out of range for axis '{axis}'")
                resolved.append(index)
        return resolved

    def _build_chunk_keys(
        self, axis_values: Mapping[str, Sequence[Any]] | None = None
    ) -> list[ChunkKey]:
        if axis_values is None:
            if self._axis_values is None:
                raise ValueError(
                    "axis_values must be set before running memoized function"
                )
            axis_values = self._axis_values
        if self.chunk_enumerator is not None:
            if self._axis_values_serializable is not None:
                return list(self.chunk_enumerator(dict(self._axis_values_serializable)))
            return list(self.chunk_enumerator(dict(axis_values)))

        axis_order, ordered = self._ordered_axis_values(axis_values)
        axis_chunks = [_chunk_values(values, size) for _, values, size in ordered]

        chunk_keys: list[ChunkKey] = []
        for product in itertools.product(*axis_chunks):
            chunk_key = tuple(
                (axis, tuple(values)) for axis, values in zip(axis_order, product)
            )
            chunk_keys.append(chunk_key)
        return chunk_keys

    def _chunk_indices_from_key(self, chunk_key: ChunkKey) -> dict[str, Any]:
        if self._axis_index_map is None:
            raise ValueError("axis_values must be set before checking cache status")
        indices: dict[str, Any] = {}
        for axis, values in chunk_key:
            axis_map = self._axis_index_map.get(axis)
            if axis_map is None:
                raise KeyError(f"Unknown axis '{axis}'")
            raw_indices = [axis_map[value] for value in values]
            indices[axis] = raw_indices
        return indices

    def _build_chunk_plan_for_axes(
        self, axis_values: Mapping[str, Sequence[Any]]
    ) -> Tuple[list[ChunkKey], dict[ChunkKey, list[Tuple[Any, ...]]]]:
        if self._axis_values is None or self._axis_index_map is None:
            raise ValueError("axis_values must be set before running memoized function")
        axis_order = self._resolve_axis_order(self._axis_values)
        per_axis_chunks: list[list[dict[str, Any]]] = []
        for axis, requested_values, size in self._ordered_axis_requests(
            axis_values, axis_order
        ):
            full_values = self._get_all_axis_values(axis)
            chunk_map: dict[int, dict[str, Any]] = {}
            for value in requested_values:
                index = self._axis_index_map[axis].get(value)
                if index is None:
                    raise KeyError(
                        f"Value '{value}' not found in axis_values for axis '{axis}'"
                    )
                chunk_id = index // size
                chunk = chunk_map.setdefault(
                    chunk_id, {"values": None, "requested": []}
                )
                chunk["requested"].append(value)
            for chunk_id, chunk in chunk_map.items():
                start = chunk_id * size
                end = min(start + size, len(full_values))
                chunk["values"] = tuple(full_values[start:end])
            per_axis_chunks.append([chunk_map[idx] for idx in sorted(chunk_map.keys())])

        chunk_keys: list[ChunkKey] = []
        requested_items: dict[ChunkKey, list[Tuple[Any, ...]]] = {}
        for combo in itertools.product(*per_axis_chunks):
            chunk_key = tuple(
                (axis, tuple(desc["values"])) for axis, desc in zip(axis_order, combo)
            )
            chunk_keys.append(chunk_key)
            requested_lists = [desc["requested"] for desc in combo]
            item_values = [
                tuple(values) for values in itertools.product(*requested_lists)
            ]
            requested_items[chunk_key] = item_values
        return chunk_keys, requested_items

    def build_item_maps_from_axis_values(
        self,
        chunk_key: ChunkKey,
        axis_values: Sequence[Tuple[Any, ...]],
        outputs: Sequence[Any],
    ) -> tuple[dict[str, Any], dict[str, dict[str, Any]]]:
        axis_names = [axis for axis, _ in chunk_key]
        item_map: dict[str, Any] = {}
        item_axis_vals: dict[str, dict[str, Any]] = {}
        for values, output in zip(axis_values, outputs):
            item_key = self.item_hash(chunk_key, values)
            item_map[item_key] = output
            item_axis_vals[item_key] = dict(zip(axis_names, values))
        return item_map, item_axis_vals

    def build_item_maps_from_chunk_output(
        self,
        chunk_key: ChunkKey,
        chunk_output: Any,
    ) -> tuple[dict[str, Any] | None, dict[str, dict[str, Any]] | None]:
        if not isinstance(chunk_output, (list, tuple)):
            return None, None
        axis_values = list(self.iter_chunk_axis_values(chunk_key))
        if len(axis_values) != len(chunk_output):
            return None, None
        return self.build_item_maps_from_axis_values(
            chunk_key,
            axis_values,
            chunk_output,
        )

    def reconstruct_output_from_items(
        self, chunk_key: ChunkKey, items: Mapping[str, Any]
    ) -> list[Any] | None:
        axis_values = list(self.iter_chunk_axis_values(chunk_key))
        if not axis_values:
            return None
        outputs: list[Any] = []
        for values in axis_values:
            item_key = self.item_hash(chunk_key, values)
            if item_key not in items:
                return None
            outputs.append(items[item_key])
        return outputs

    def reconstruct_partial_output_from_items(
        self,
        chunk_key: ChunkKey,
        items: Mapping[str, Any],
    ) -> list[Any]:
        axis_values = list(self.iter_chunk_axis_values(chunk_key))
        outputs: list[Any] = []
        for values in axis_values:
            item_key = self.item_hash(chunk_key, values)
            if item_key in items:
                outputs.append(items[item_key])
        return outputs

    def extract_items_from_map(
        self,
        item_map: Mapping[str, Any] | None,
        chunk_key: ChunkKey,
        requested_items: list[Tuple[Any, ...]],
    ) -> list[Any] | None:
        if item_map is None:
            return None
        outputs: list[Any] = []
        for values in requested_items:
            item_key = self.item_hash(chunk_key, values)
            if item_key not in item_map:
                return None
            outputs.append(item_map[item_key])
        return outputs

    def item_hash(self, chunk_key: ChunkKey, axis_values: Tuple[Any, ...]) -> str:
        payload = {
            "axis_values": tuple(
                (axis, value) for (axis, _), value in zip(chunk_key, axis_values)
            ),
            "version": self.version,
        }
        data = stable_serialize(payload)
        return hashlib.sha256(data.encode("utf-8")).hexdigest()

    def iter_chunk_axis_values(self, chunk_key: ChunkKey) -> Sequence[Tuple[Any, ...]]:
        axis_values = [values for _, values in chunk_key]
        return list(itertools.product(*axis_values))

    def load_payload(self, path: Path) -> dict[str, Any] | None:
        if not path.exists():
            return None
        with open(path, "rb") as handle:
            payload = pickle.load(handle)
        return cast(dict[str, Any], payload)

    def collect_chunk_data(
        self,
        payload: Mapping[str, Any],
        chunk_key: ChunkKey,
        requested_items: list[Tuple[Any, ...]] | None,
        collate_fn: Callable[[list[Any]], Any],
    ) -> tuple[Any | None, bool]:
        if requested_items is None:
            chunk_output = payload.get("output")
            if chunk_output is None:
                items = payload.get("items")
                if items is not None:
                    chunk_output = self.reconstruct_output_from_items(chunk_key, items)
                    if chunk_output is None:
                        return (
                            self.reconstruct_partial_output_from_items(
                                chunk_key, items
                            ),
                            True,
                        )
            if chunk_output is None:
                return None, False
            return chunk_output, False
        item_map = payload.get("items")
        if item_map is None:
            return None, False
        cached_outputs: list[Any] = []
        missing = False
        for values in requested_items:
            item_key = self.item_hash(chunk_key, values)
            if item_key not in item_map:
                missing = True
                continue
            cached_outputs.append(item_map[item_key])
        if not cached_outputs:
            return None, False
        return collate_fn([cached_outputs]), missing

    def _materialize_axis_values(self, axis_values_obj: Any) -> list[Any]:
        if isinstance(axis_values_obj, ABCSequence) and not isinstance(
            axis_values_obj, (str, bytes, bytearray)
        ):
            return list(axis_values_obj)
        values = list(axis_values_obj)
        return sorted(values, key=stable_serialize)

    def _resolve_axis_chunk_size(self, axis: str) -> int:
        spec = self.chunk_spec.get(axis)
        if spec is None:
            raise KeyError(f"Missing chunk_spec for axis '{axis}'")
        if isinstance(spec, dict):
            size = spec.get("size")
            if size is None:
                raise KeyError(f"Missing size for axis '{axis}'")
            return int(size)
        return int(spec)

    @classmethod
    def discover_caches(cls, root: str | Path) -> list[dict[str, Any]]:
        """Discover all existing caches in root.

        Returns a list of cache metadata dictionaries, each containing:
        - cache_id: The identity string of the cache
        - path: Path to the cache directory
        - metadata: Full metadata dict if available, None otherwise

        Args:
            root: Directory to scan for caches.
        """
        root = Path(root)
        if not root.exists():
            return []
        caches: list[dict[str, Any]] = []
        for entry in root.iterdir():
            if not entry.is_dir():
                continue
            metadata_path = entry / "metadata.json"
            metadata = cls._read_metadata_file(metadata_path, required=False)
            if metadata is None:
                continue
            caches.append(
                {
                    "cache_id": entry.name,
                    "path": entry,
                    "metadata": metadata,
                }
            )
        return caches

    @classmethod
    def _is_superset_compatible(
        cls,
        cache_metadata: dict[str, Any] | None,
        req_axis_values: dict[str, Any] | None,
    ) -> bool:
        """Check if an existing cache is a superset of requested axis values."""
        if cache_metadata is None or req_axis_values is None:
            return False
        meta_axis_values = cache_metadata.get("axis_values", {})
        req_axis_set = set(req_axis_values.keys())
        meta_axis_set = set(meta_axis_values.keys())

        if not req_axis_set.issubset(meta_axis_set):
            return False

        for axis, req_value in req_axis_values.items():
            meta_vals = cls._axis_value_set(meta_axis_values[axis])
            req_vals = cls._axis_value_set(req_value)
            if not req_vals.issubset(meta_vals):
                return False

        return True

    @classmethod
    def cache_is_compatible(
        cls,
        cache,
        cache_id: str | None = None,
        axis_values: dict[str, Any] | None = None,
        chunk_spec: dict[str, Any] | None = None,
        version: str | None = None,
        axis_order: Sequence[str] | None = None,
        allow_superset: bool = False,
    ) -> bool:
        """Find caches compatible with the given criteria.

        Matching rules:
        - If a criterion is provided, it must match exactly.
        - If a criterion is None (omitted), it is ignored (wildcard).
        - If allow_superset is True with axis_values provided,
          also finds caches that are supersets (contain all requested data).

        Args:
            cache: Cache.
            cache_id: Optional cache_id to match.
            axis_values: Optional axis_values dict to match.
            chunk_spec: Optional chunk_spec dict to match.
            version: Optional version string to match.
            axis_order: Optional axis_order sequence to match.
            allow_superset: If True, find superset caches when axis_values
                are provided.
        """
        metadata = cache.get("metadata")

        if allow_superset:
            return cls._is_superset_compatible(metadata, axis_values)

        if metadata is None:
            return False
        if cache_id is not None:
            if cache.get("cache_id") != cache_id:
                return False
        if axis_values is not None:
            meta_axis_values = metadata.get("axis_values", {})
            if stable_serialize(meta_axis_values) != stable_serialize(axis_values):
                return False
        if chunk_spec is not None:
            meta_spec = metadata.get("chunk_spec", {})
            if stable_serialize(meta_spec) != stable_serialize(chunk_spec):
                return False
        if version is not None:
            if metadata.get("version") != version:
                return False
        if axis_order is not None:
            meta_axis_order = metadata.get("axis_order")
            meta_tuple = tuple(meta_axis_order) if meta_axis_order is not None else None
            if meta_tuple != tuple(axis_order):
                return False
        return True

    @classmethod
    def find_compatible_caches(
        cls,
        root: str | Path,
        cache_id: str | None = None,
        axis_values: dict[str, Any] | None = None,
        chunk_spec: dict[str, Any] | None = None,
        version: str | None = None,
        axis_order: Sequence[str] | None = None,
        allow_superset: bool = False,
    ) -> list[dict[str, Any]]:
        """Find caches compatible with the given criteria.

        Matching rules:
        - If a criterion is provided, it must match exactly.
        - If a criterion is None (omitted), it is ignored (wildcard).
        - If allow_superset is True with axis_values provided,
          also finds caches that are supersets (contain all requested data).

        Returns a list of compatible cache entries, each containing:
        - cache_id: The identity string of the cache
        - path: Path to the cache directory
        - metadata: Full metadata dict

        Args:
            root: Directory to scan for caches.
            cache_id: Optional cache_id to match.
            axis_values: Optional axis_values dict to match.
            chunk_spec: Optional chunk_spec dict to match.
            version: Optional version string to match.
            axis_order: Optional axis_order sequence to match.
            allow_superset: If True, find superset caches when axis_values
                are provided.
        """
        all_caches = cls.discover_caches(root)
        compatible: list[dict[str, Any]] = []

        for cache in all_caches:
            if cls.cache_is_compatible(
                cache,
                cache_id,
                axis_values,
                chunk_spec,
                version,
                axis_order,
                allow_superset,
            ):
                compatible.append(cache)

        return compatible

    @classmethod
    def load_from_cache(
        cls,
        root: str | Path,
        cache_id: str,
        collate_fn: CollateFn | None = None,
        chunk_enumerator: MemoChunkEnumerator | None = None,
        chunk_hash_fn: Callable[[str, ChunkKey, str], str] | None = None,
        path_fn: CachePathFn | None = None,
        verbose: int = 1,
        profile: bool = False,
        exclusive: bool = False,
        warn_on_overlap: bool = False,
    ) -> "ChunkCache":
        """Load a ChunkCache instance from an existing cache by cache_id.

        Raises:
            FileNotFoundError: If the cache directory or metadata.json does not exist.
            ValueError: If metadata is invalid.

        Args:
            root: Directory containing caches.
            cache_id: The identity string of the specific cache.
            collate_fn: Optional collate function for the list of chunk outputs.
            chunk_enumerator: Optional chunk enumerator.
            chunk_hash_fn: Optional override for chunk hashing.
            path_fn: Optional override for cache file paths.
            verbose: Verbosity flag.
            profile: Enable profiling output.
            exclusive: If True, error when creating a cache with conflicting
                metadata for the same cache_id.
            warn_on_overlap: If True, warn when caches overlap.
        """
        root = Path(root)
        cache_path = root / cache_id
        metadata_path = cache_path / "metadata.json"
        if not cache_path.exists():
            raise FileNotFoundError(f"Cache directory not found: {cache_path}")
        metadata = cls._read_metadata_file(metadata_path, required=True)
        if metadata is None:
            raise ValueError(f"Invalid metadata in {metadata_path}")
        axis_values = metadata.get("axis_values")
        if axis_values is None:
            raise ValueError("Metadata missing 'axis_values'")
        chunk_spec = metadata.get("chunk_spec")
        if chunk_spec is None:
            raise ValueError("Metadata missing 'chunk_spec'")
        instance = cls(
            root=cache_path,
            chunk_spec=chunk_spec,
            axis_values=axis_values,
            cache_id=metadata.get("cache_id", cache_id),
            metadata=metadata.get("metadata", {}),
            collate_fn=collate_fn,
            chunk_enumerator=chunk_enumerator,
            chunk_hash_fn=chunk_hash_fn,
            path_fn=path_fn,
            version=metadata.get("version", "v1"),
            axis_order=metadata.get("axis_order"),
            verbose=verbose,
            profile=profile,
            exclusive=exclusive,
            warn_on_overlap=warn_on_overlap,
        )
        return instance

    @staticmethod
    def _axis_value_set(values: Any) -> set[Any]:
        if isinstance(values, (list, tuple)):
            return set(values)
        return {values}

    @staticmethod
    def _read_metadata_file(
        metadata_path: Path, *, required: bool
    ) -> dict[str, Any] | None:
        if not metadata_path.exists():
            if required:
                raise FileNotFoundError(f"Metadata file not found: {metadata_path}")
            return None
        try:
            with open(metadata_path, "r", encoding="utf-8") as handle:
                return json.load(handle)
        except (json.JSONDecodeError, IOError) as exc:
            if required:
                raise ValueError(f"Invalid metadata in {metadata_path}") from exc
            return None



import hashlib
from typing import Any, Mapping


def stable_serialize(value: Any) -> str:
    if isinstance(value, dict):
        items = ((k, stable_serialize(value[k])) for k in sorted(value))
        return "{" + ",".join(f"{k}:{v}" for k, v in items) + "}"
    if isinstance(value, (list, tuple)):
        inner = ",".join(stable_serialize(v) for v in value)
        return "[" + inner + "]"
    return repr(value)


def params_to_cache_id(params: Mapping[str, Any]) -> str:
    data = stable_serialize(dict(params))
    return hashlib.sha256(data.encode("utf-8")).hexdigest()



from __future__ import annotations

from typing import Any, Callable, Mapping, Sequence, Tuple, cast

from ._format import chunk_key_size, prepare_progress, print_chunk_summary
from .runner_protocol import (
    BuildItemMapsFromChunkOutputFn,
    CacheProtocol,
    ChunkHashFn,
    ResolveCachePathFn,
    RunnerContext,
    UpdateChunkIndexFn,
    WriteMetadataFn,
    WriteChunkPayloadFn,
    CollectChunkDataFn,
    ExtractItemsFromMapFn,
    LoadPayloadFn,
)
from .runners_parallel import run_parallel
from .runners_common import (
    ChunkKey,
    Diagnostics,
    _log_chunk,
    _merge_outputs,
    _payload_item_map,
    _stream_item_count,
    resolve_chunk_path,
    resolve_cache_for_run,
    resolve_runner_deps,
)

MergeFn = Callable[[list[Any]], Any]


def run(
    cache: Any,
    exec_fn: Callable[..., Any],
    *,
    collate_fn: Callable[[list[Any]], Any] | None = None,
    params: dict[str, Any] | None = None,
    axis_values_override: Mapping[str, Sequence[Any]] | None = None,
    extend_cache: bool = False,
    allow_superset: bool = False,
    # Manual cache methods (optional overrides)
    write_metadata: WriteMetadataFn | None = None,
    chunk_hash: ChunkHashFn | None = None,
    resolve_cache_path: ResolveCachePathFn | None = None,
    load_payload: LoadPayloadFn | None = None,
    write_chunk_payload: WriteChunkPayloadFn | None = None,
    update_chunk_index: UpdateChunkIndexFn | None = None,
    build_item_maps_from_chunk_output: BuildItemMapsFromChunkOutputFn | None = None,
    collect_chunk_data: CollectChunkDataFn | None = None,
    extract_items_from_map: ExtractItemsFromMapFn | None = None,
    context: RunnerContext | None = None,
) -> Tuple[Any, Diagnostics]:
    """Run memoized execution with output via the cache runner.

    The cache must already represent the desired axis subset.
    collate_fn overrides cache.collate_fn for this run.
    Partial chunks return whatever cached output is available and increment
    diagnostics.partial_chunks.
    """
    cache = resolve_cache_for_run(
        cache,
        params=params,
        axis_values_override=axis_values_override,
        extend_cache=extend_cache,
        allow_superset=allow_superset,
    )
    exec_fn_bound = cache.bind_exec_fn(exec_fn)
    write_metadata_fn = (
        write_metadata if write_metadata is not None else cache.write_metadata
    )
    write_metadata_fn()
    chunk_keys = cache.resolved_chunk_keys()
    return run_chunks(
        chunk_keys,
        exec_fn_bound,
        cache=cache,
        collate_fn=collate_fn,
        chunk_hash=chunk_hash,
        resolve_cache_path=resolve_cache_path,
        load_payload=load_payload,
        write_chunk_payload=write_chunk_payload,
        update_chunk_index=update_chunk_index,
        build_item_maps_from_chunk_output=build_item_maps_from_chunk_output,
        collect_chunk_data=collect_chunk_data,
        extract_items_from_map=extract_items_from_map,
        context=context,
    )


def run_streaming(
    cache: Any,
    exec_fn: Callable[..., Any],
    *,
    collate_fn: Callable[[list[Any]], Any] | None = None,
    params: dict[str, Any] | None = None,
    axis_values_override: Mapping[str, Sequence[Any]] | None = None,
    extend_cache: bool = False,
    allow_superset: bool = False,
    # Manual cache methods (optional overrides)
    write_metadata: WriteMetadataFn | None = None,
    chunk_hash: ChunkHashFn | None = None,
    resolve_cache_path: ResolveCachePathFn | None = None,
    load_payload: LoadPayloadFn | None = None,
    write_chunk_payload: WriteChunkPayloadFn | None = None,
    update_chunk_index: UpdateChunkIndexFn | None = None,
    build_item_maps_from_chunk_output: BuildItemMapsFromChunkOutputFn | None = None,
    collect_chunk_data: CollectChunkDataFn | None = None,
    extract_items_from_map: ExtractItemsFromMapFn | None = None,
    context: RunnerContext | None = None,
) -> Diagnostics:
    """Run memoized execution without returning outputs.

    The cache must already represent the desired axis subset.
    collate_fn is accepted for API parity but has no effect without outputs.
    Partial chunks still count as cached and increment diagnostics.partial_chunks.
    """
    cache = resolve_cache_for_run(
        cache,
        params=params,
        axis_values_override=axis_values_override,
        extend_cache=extend_cache,
        allow_superset=allow_superset,
    )
    exec_fn_bound = cache.bind_exec_fn(exec_fn)
    write_metadata_fn = (
        write_metadata if write_metadata is not None else cache.write_metadata
    )
    write_metadata_fn()
    chunk_keys = cache.resolved_chunk_keys()
    return run_chunks_streaming(
        chunk_keys,
        exec_fn_bound,
        cache=cache,
        collate_fn=collate_fn,
        chunk_hash=chunk_hash,
        resolve_cache_path=resolve_cache_path,
        load_payload=load_payload,
        write_chunk_payload=write_chunk_payload,
        update_chunk_index=update_chunk_index,
        build_item_maps_from_chunk_output=build_item_maps_from_chunk_output,
        collect_chunk_data=collect_chunk_data,
        extract_items_from_map=extract_items_from_map,
        context=context,
    )


def execute_and_save_chunk(
    chunk_key: ChunkKey,
    exec_fn: Callable[..., Any],
    chunk_hash: str,
    diagnostics: Diagnostics,
    *,
    resolve_cache_path: ResolveCachePathFn,
    write_chunk_payload: WriteChunkPayloadFn,
    update_chunk_index: UpdateChunkIndexFn,
    build_item_maps_from_chunk_output: BuildItemMapsFromChunkOutputFn,
    existing_payload: Mapping[str, Any] | None = None,
) -> tuple[Any, dict[str, Any] | None]:
    """Execute a chunk, persist payload, and update index."""
    diagnostics.executed_chunks += 1
    chunk_axes = {axis: list(values) for axis, values in chunk_key}
    chunk_output = exec_fn(**chunk_axes)
    diagnostics.max_stream_items = max(
        diagnostics.max_stream_items,
        _stream_item_count(chunk_output),
    )
    payload: dict[str, Any] = {}
    item_map, item_axis_vals = build_item_maps_from_chunk_output(
        chunk_key,
        chunk_output=chunk_output,
    )
    if item_map is not None:
        payload["items"] = item_map
        if item_axis_vals is not None:
            payload["axis_vals"] = item_axis_vals
    else:
        payload["output"] = chunk_output
    path = resolve_cache_path(chunk_key, chunk_hash)
    write_chunk_payload(
        path,
        payload,
        existing=existing_payload,
    )
    update_chunk_index(chunk_hash, chunk_key)
    return chunk_output, item_map


def run_chunks(
    chunk_keys: Sequence[ChunkKey],
    exec_fn: Callable[..., Any],
    *,
    cache: CacheProtocol,
    collate_fn: Callable[[list[Any]], Any] | None = None,
    # Manual cache methods (optional overrides)
    chunk_hash: ChunkHashFn | None = None,
    resolve_cache_path: ResolveCachePathFn | None = None,
    load_payload: LoadPayloadFn | None = None,
    write_chunk_payload: WriteChunkPayloadFn | None = None,
    update_chunk_index: UpdateChunkIndexFn | None = None,
    build_item_maps_from_chunk_output: BuildItemMapsFromChunkOutputFn | None = None,
    collect_chunk_data: CollectChunkDataFn | None = None,
    extract_items_from_map: ExtractItemsFromMapFn | None = None,
    context: RunnerContext | None = None,
) -> Tuple[Any, Diagnostics]:
    """Run a list of chunk keys and return merged output.

    collate_fn overrides cache.collate_fn for this run.
    Partial chunks return whatever cached output is available and increment
    diagnostics.partial_chunks.
    """
    deps = resolve_runner_deps(
        cache=cache,
        context=context,
        chunk_hash=chunk_hash,
        resolve_cache_path=resolve_cache_path,
        load_payload=load_payload,
        write_chunk_payload=write_chunk_payload,
        update_chunk_index=update_chunk_index,
        build_item_maps_from_chunk_output=build_item_maps_from_chunk_output,
        collect_chunk_data=collect_chunk_data,
        extract_items_from_map=extract_items_from_map,
        require=[
            "context",
            "chunk_hash",
            "resolve_cache_path",
            "load_payload",
            "write_chunk_payload",
            "update_chunk_index",
            "build_item_maps_from_chunk_output",
            "collect_chunk_data",
            "extract_items_from_map",
        ],
    )
    context = cast(RunnerContext, deps.context)
    collect_chunk_data = cast(CollectChunkDataFn, deps.collect_chunk_data)
    extract_items_from_map = cast(ExtractItemsFromMapFn, deps.extract_items_from_map)
    requested_items_map = cache.requested_items_by_chunk()
    collate_fn_resolved: MergeFn
    if collate_fn is not None:
        collate_fn_resolved = collate_fn
    elif context.collate_fn is not None:
        collate_fn_resolved = context.collate_fn
    else:
        collate_fn_resolved = lambda chunk: chunk
    diagnostics = Diagnostics(total_chunks=len(chunk_keys))
    total_chunks = len(chunk_keys)
    total_items = sum(chunk_key_size(chunk_key) for chunk_key in chunk_keys)
    report_progress, update_processed = prepare_progress(
        total_chunks=total_chunks,
        total_items=total_items,
        verbose=context.verbose,
        label="planning",
    )

    def process_chunk(chunk_key: ChunkKey) -> tuple[Any, bool, bool]:
        chunk_hash_value, path = resolve_chunk_path(
            cast(ChunkHashFn, deps.chunk_hash),
            cast(ResolveCachePathFn, deps.resolve_cache_path),
            chunk_key,
        )
        payload = cast(LoadPayloadFn, deps.load_payload)(path)
        existing_payload = payload
        requested_items = (
            requested_items_map.get(chunk_key)
            if requested_items_map is not None
            else None
        )
        if payload is not None:
            cached_output, is_partial = collect_chunk_data(
                payload,
                chunk_key,
                requested_items,
                collate_fn_resolved,
            )
            if cached_output is not None:
                _log_chunk(
                    context,
                    "load",
                    chunk_key,
                    None if requested_items is None else len(requested_items),
                )
                return cached_output, True, is_partial

        chunk_output, item_map = execute_and_save_chunk(
            chunk_key,
            exec_fn,
            chunk_hash_value,
            diagnostics,
            resolve_cache_path=cast(ResolveCachePathFn, deps.resolve_cache_path),
            write_chunk_payload=cast(WriteChunkPayloadFn, deps.write_chunk_payload),
            update_chunk_index=cast(UpdateChunkIndexFn, deps.update_chunk_index),
            build_item_maps_from_chunk_output=cast(
                BuildItemMapsFromChunkOutputFn, deps.build_item_maps_from_chunk_output
            ),
            existing_payload=existing_payload,
        )

        if requested_items is None:
            _log_chunk(context, "run", chunk_key, None)
            return chunk_output, False, False
        _log_chunk(context, "run", chunk_key, len(requested_items))
        extracted = extract_items_from_map(
            item_map,
            chunk_key,
            requested_items,
        )
        result = (
            collate_fn_resolved([extracted]) if extracted is not None else chunk_output
        )
        return result, False, False

    outputs: list[Any] = []
    for processed, chunk_key in enumerate(chunk_keys, start=1):
        update_processed(chunk_key_size(chunk_key))
        output, cached, is_partial = process_chunk(chunk_key)
        if cached:
            diagnostics.cached_chunks += 1
        if is_partial:
            diagnostics.partial_chunks += 1
        outputs.append(output)
        report_progress(processed, processed == total_chunks)

    merged = _merge_outputs(
        context,
        outputs,
        diagnostics,
        collate_fn=collate_fn_resolved,
    )
    print_chunk_summary(diagnostics, context.verbose)
    return merged, diagnostics


def run_chunks_streaming(
    chunk_keys: Sequence[ChunkKey],
    exec_fn: Callable[..., Any],
    *,
    cache: CacheProtocol,
    collate_fn: Callable[[list[Any]], Any] | None = None,
    # Manual cache methods (optional overrides)
    chunk_hash: ChunkHashFn | None = None,
    resolve_cache_path: ResolveCachePathFn | None = None,
    load_payload: LoadPayloadFn | None = None,
    write_chunk_payload: WriteChunkPayloadFn | None = None,
    update_chunk_index: UpdateChunkIndexFn | None = None,
    build_item_maps_from_chunk_output: BuildItemMapsFromChunkOutputFn | None = None,
    collect_chunk_data: CollectChunkDataFn | None = None,
    extract_items_from_map: ExtractItemsFromMapFn | None = None,
    context: RunnerContext | None = None,
) -> Diagnostics:
    """Run chunks and flush payloads to disk only.

    collate_fn is accepted for API parity but has no effect without outputs.
    Partial chunks still count as cached and increment diagnostics.partial_chunks.
    """
    deps = resolve_runner_deps(
        cache=cache,
        context=context,
        chunk_hash=chunk_hash,
        resolve_cache_path=resolve_cache_path,
        load_payload=load_payload,
        write_chunk_payload=write_chunk_payload,
        update_chunk_index=update_chunk_index,
        build_item_maps_from_chunk_output=build_item_maps_from_chunk_output,
        collect_chunk_data=collect_chunk_data,
        extract_items_from_map=extract_items_from_map,
        require=[
            "context",
            "chunk_hash",
            "resolve_cache_path",
            "load_payload",
            "write_chunk_payload",
            "update_chunk_index",
            "build_item_maps_from_chunk_output",
            "collect_chunk_data",
            "extract_items_from_map",
        ],
    )
    context = cast(RunnerContext, deps.context)
    requested_items_map = cache.requested_items_by_chunk()
    build_item_maps_from_chunk_output = cast(
        BuildItemMapsFromChunkOutputFn, deps.build_item_maps_from_chunk_output
    )
    diagnostics = Diagnostics(total_chunks=len(chunk_keys))
    total_chunks = len(chunk_keys)
    total_items = sum(chunk_key_size(chunk_key) for chunk_key in chunk_keys)
    report_progress, update_processed = prepare_progress(
        total_chunks=total_chunks,
        total_items=total_items,
        verbose=context.verbose,
        label="planning",
    )

    for processed, chunk_key in enumerate(chunk_keys, start=1):
        update_processed(chunk_key_size(chunk_key))
        chunk_hash_value, path = resolve_chunk_path(
            cast(ChunkHashFn, deps.chunk_hash),
            cast(ResolveCachePathFn, deps.resolve_cache_path),
            chunk_key,
        )
        payload = cast(LoadPayloadFn, deps.load_payload)(path)
        if payload is not None:
            if requested_items_map is None:
                items_payload = payload.get("items")
                if payload.get("output") is None and items_payload is not None:
                    if (
                        cache.reconstruct_output_from_items(chunk_key, items_payload)
                        is None
                    ):
                        diagnostics.partial_chunks += 1
                diagnostics.cached_chunks += 1
                _log_chunk(context, "load", chunk_key, None)
                report_progress(processed, processed == total_chunks)
                continue
            item_map = _payload_item_map(
                build_item_maps_from_chunk_output=build_item_maps_from_chunk_output,
                resolve_cache_path=cast(ResolveCachePathFn, deps.resolve_cache_path),
                write_chunk_payload=cast(WriteChunkPayloadFn, deps.write_chunk_payload),
                chunk_key=chunk_key,
                payload=payload,
                chunk_hash=chunk_hash_value,
                write_back=True,
            )
            if item_map is not None:
                diagnostics.cached_chunks += 1
                requested_items = requested_items_map.get(chunk_key)
                if requested_items is not None:
                    for values in requested_items:
                        item_key = cache.item_hash(chunk_key, values)
                        if item_key not in item_map:
                            diagnostics.partial_chunks += 1
                            break
                _log_chunk(
                    context,
                    "load",
                    chunk_key,
                    None if requested_items is None else len(requested_items),
                )
                report_progress(processed, processed == total_chunks)
                continue

        execute_and_save_chunk(
            chunk_key,
            exec_fn,
            chunk_hash_value,
            diagnostics,
            resolve_cache_path=cast(ResolveCachePathFn, deps.resolve_cache_path),
            write_chunk_payload=cast(WriteChunkPayloadFn, deps.write_chunk_payload),
            update_chunk_index=cast(UpdateChunkIndexFn, deps.update_chunk_index),
            build_item_maps_from_chunk_output=cast(
                BuildItemMapsFromChunkOutputFn, deps.build_item_maps_from_chunk_output
            ),
            existing_payload=None,
        )

        requested_items = (
            requested_items_map.get(chunk_key)
            if requested_items_map is not None
            else None
        )
        _log_chunk(
            context,
            "run",
            chunk_key,
            None if requested_items is None else len(requested_items),
        )

        report_progress(processed, processed == total_chunks)

    print_chunk_summary(diagnostics, context.verbose)
    return diagnostics



import math
import threading
import time
from dataclasses import dataclass
from typing import Any, Callable, Mapping, Sequence, Tuple

ChunkKey = Tuple[Tuple[str, Tuple[Any, ...]], ...]

_thread_local = threading.local()
PROGRESS_REPORT_DIVISOR = 50


@dataclass
class _ProgressTracker:
    last_len: int = 0
    last_msg: str = ""

    def print_progress(self, message: str, *, final: bool) -> None:
        if final:
            print(message, end="\n", flush=True)
            self.last_len = 0
            self.last_msg = ""
            return
        pad = max(self.last_len - len(message), 0)
        print(message + (" " * pad), end="\r", flush=True)
        self.last_len = len(message)
        self.last_msg = message

    def print_detail(self, message: str) -> None:
        if self.last_len:
            print()
        print(message)
        if self.last_msg:
            print(self.last_msg, end="\r", flush=True)
            self.last_len = len(self.last_msg)


def format_axis_values(values: Any) -> str:
    if isinstance(values, (list, tuple)):
        if len(values) <= 4:
            inner = ", ".join(repr(value) for value in values)
            return f"[{inner}]"
        head = ", ".join(repr(value) for value in values[:2])
        tail = ", ".join(repr(value) for value in values[-2:])
        return f"[{head}, ..., {tail}]"
    return repr(values)


def format_cache_id(cache_id: str | None) -> list[str]:
    lines = ["[ChunkCache] cache_id:"]
    if not cache_id:
        lines.append("  (none)")
        return lines
    lines.append(f"  {cache_id}")
    return lines


def format_params(params: Mapping[str, Any]) -> list[str]:
    lines = ["[ChunkCache] params:"]
    if not params:
        lines.append("  (none)")
        return lines
    for key in sorted(params):
        value = params[key]
        lines.append(f"  {key}={value!r}")
    return lines


def format_spec(axis_values: Mapping[str, Any], axis_order: Sequence[str]) -> list[str]:
    lines = ["[ChunkCache] spec:"]
    for axis in axis_order:
        values = axis_values.get(axis)
        lines.append(f"  {axis}={format_axis_values(values)}")
    return lines


def format_eta(seconds: float) -> str:
    if seconds < 0 or not math.isfinite(seconds):
        return "--:--:--"
    total = int(seconds)
    hours = total // 3600
    minutes = (total % 3600) // 60
    secs = total % 60
    return f"{hours:02d}:{minutes:02d}:{secs:02d}"


def format_rate_eta(
    label: str,
    processed: int,
    total: int,
    start_time: float,
    *,
    rate_processed: int | None = None,
    rate_total: int | None = None,
) -> str:
    elapsed = max(time.monotonic() - start_time, 1e-6)
    if rate_processed is None:
        rate_processed = processed
    if rate_total is None:
        rate_total = total
    rate = rate_processed / elapsed
    percent = 100.0 if total <= 0 else (processed / total * 100.0)
    remaining = 0 if rate_total <= 0 else max(rate_total - rate_processed, 0)
    eta = remaining / rate if rate > 0 else float("inf")
    return (
        f"[ChunkCache] {label} {processed}/{total} "
        f"({percent:0.1f}%) rate={rate:0.1f}/s ETA={format_eta(eta)}"
    )


def print_progress(message: str, *, final: bool) -> None:
    if not hasattr(_thread_local, "tracker"):
        _thread_local.tracker = _ProgressTracker()
    _thread_local.tracker.print_progress(message, final=final)


def print_detail(message: str) -> None:
    if not hasattr(_thread_local, "tracker"):
        _thread_local.tracker = _ProgressTracker()
    _thread_local.tracker.print_detail(message)


def chunk_key_size(chunk_key: ChunkKey) -> int:
    if not chunk_key:
        return 0
    return math.prod(len(values) for _, values in chunk_key)


def build_plan_lines(
    cache_id: str | None,
    params: Mapping[str, Any],
    axis_values: Mapping[str, Any],
    axis_order: Sequence[str],
    cached_count: int,
    execute_count: int,
) -> list[str]:
    lines: list[str] = []
    lines.extend(format_cache_id(cache_id))
    lines.extend(format_params(params))
    lines.extend(format_spec(axis_values, axis_order))
    lines.append(f"[ChunkCache] plan: cached={cached_count} execute={execute_count}")
    return lines


def prepare_progress(
    *,
    total_chunks: int,
    total_items: int,
    verbose: int,
    label: str = "planning",
) -> tuple[Callable[[int, bool], None], Callable[[int], None]]:
    progress_step = max(1, total_chunks // PROGRESS_REPORT_DIVISOR)
    start_time = time.monotonic()
    processed_items = 0

    def update_processed(count: int) -> None:
        nonlocal processed_items
        processed_items += count

    def report_progress(processed: int, final: bool = False) -> None:
        if verbose != 1:
            return
        if not final and processed % progress_step != 0 and processed != total_chunks:
            return
        message = format_rate_eta(
            label,
            processed_items,
            total_items,
            start_time,
        )
        print_progress(message, final=final)

    return report_progress, update_processed


def print_chunk_summary(
    diagnostics: Any,
    verbose: int,
) -> None:
    if verbose >= 2:
        partial_chunks = getattr(diagnostics, "partial_chunks", 0)
        partial_suffix = f" partial={partial_chunks}" if partial_chunks else ""
        print_detail(
            "[ChunkCache] summary "
            f"cached={diagnostics.cached_chunks} "
            f"executed={diagnostics.executed_chunks} "
            f"total={diagnostics.total_chunks}" + partial_suffix
        )



from __future__ import annotations

import dataclasses
import functools
import inspect
import time
from bisect import bisect_right
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Any, Callable, Iterable, Mapping, Sequence, Tuple, cast

from ._format import (
    build_plan_lines,
    chunk_key_size,
    format_rate_eta,
    print_chunk_summary,
    print_progress,
)
from .runner_protocol import (
    BuildItemMapsFromAxisValuesFn,
    BuildItemMapsFromChunkOutputFn,
    CacheProtocol,
    CacheStatus,
    ChunkHashFn,
    CollectChunkDataFn,
    ItemHashFn,
    LoadChunkIndexFn,
    LoadPayloadFn,
    ReconstructPartialOutputFromItemsFn,
    ReconstructOutputFromItemsFn,
    ResolveCachePathFn,
    RunnerContext,
    UpdateChunkIndexFn,
    WriteChunkPayloadFn,
    WriteMetadataFn,
)
from .runners_common import (
    ChunkKey,
    Diagnostics,
    _log_chunk,
    _merge_outputs,
    _payload_item_map,
    _require_axis_info,
    _require_params,
    _save_chunk_payload,
    prepare_planning_progress,
    resolve_cache_for_run,
    resolve_runner_deps,
    resolve_chunk_path,
)

EXEC_REPORT_INTERVAL_SECONDS = 2.0


@dataclasses.dataclass
class _ParallelSetup:
    map_fn: Callable[..., Iterable[Any]]
    cached_chunks: list[ChunkKey]
    missing_chunks: list[ChunkKey]
    diagnostics: Diagnostics
    collate_fn: Callable[[list[Any]], Any]
    map_fn_kwargs: Mapping[str, Any]


@dataclasses.dataclass
class _MissingTracker:
    missing_items: list[Any] = dataclasses.field(default_factory=list)
    missing_items_by_chunk: dict[ChunkKey, list[Any]] = dataclasses.field(
        default_factory=dict
    )
    missing_chunk_order: list[ChunkKey] = dataclasses.field(default_factory=list)
    missing_item_keys: list[ChunkKey] | None = None

    @classmethod
    def create(cls, *, track_item_keys: bool) -> "_MissingTracker":
        missing_item_keys = [] if track_item_keys else None
        return cls(missing_item_keys=missing_item_keys)

    def register(self, chunk_key: ChunkKey, items_to_add: list[Any]) -> None:
        if not items_to_add:
            return
        if chunk_key not in self.missing_items_by_chunk:
            self.missing_items_by_chunk[chunk_key] = []
            self.missing_chunk_order.append(chunk_key)
        self.missing_items_by_chunk[chunk_key].extend(items_to_add)
        self.missing_items.extend(items_to_add)
        if self.missing_item_keys is not None:
            self.missing_item_keys.extend([chunk_key] * len(items_to_add))


def _exec_with_item(
    exec_fn: Callable[..., Any],
    item: Any,
) -> Any:
    if isinstance(item, Mapping):
        return exec_fn(**item)
    return exec_fn(item)


def _map_executor(
    func: Callable[..., Any],
    items: Iterable[Any],
    **kwargs: Any,
) -> Iterable[Any]:
    max_workers = kwargs.get("max_workers")
    chunksize = kwargs.get("chunksize")
    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        if chunksize is None:
            yield from executor.map(func, items)
            return
        yield from executor.map(func, items, chunksize=chunksize)


def _prepare_parallel_setup(
    context: RunnerContext,
    write_metadata_fn: WriteMetadataFn,
    cache_status: CacheStatus,
    *,
    map_fn: Callable[..., Iterable[Any]] | None,
    map_fn_kwargs: Mapping[str, Any] | None,
    collate_fn: Callable[[list[Any]], Any] | None,
    params_dict: dict[str, Any] | None = None,
    axis_values_override: Mapping[str, Any] | None = None,
    write_metadata: bool = True,
) -> _ParallelSetup:
    if map_fn is None:
        if context.verbose == 1:
            map_fn = _map_executor
        else:
            map_fn = lambda func, items, **kwargs: list(
                _map_executor(func, items, **kwargs)
            )
    map_fn = cast(Callable[..., Iterable[Any]], map_fn)

    if params_dict is None:
        params_dict = _require_params(cache_status)
    if write_metadata:
        write_metadata_fn()

    cached_chunks: list[ChunkKey] = list(cache_status.get("cached_chunks", []))
    missing_chunks: list[ChunkKey] = list(cache_status.get("missing_chunks", []))

    diagnostics = Diagnostics(total_chunks=len(cached_chunks) + len(missing_chunks))

    if collate_fn is None:
        collate_fn = (
            context.collate_fn
            if context.collate_fn is not None
            else lambda chunk: chunk
        )
    collate_fn = cast(Callable[[list[Any]], Any], collate_fn)

    if map_fn_kwargs is None:
        map_fn_kwargs = {}

    if context.verbose == 1:
        axis_values = axis_values_override or cache_status.get("axis_values")
        if isinstance(axis_values, Mapping):
            axis_order, _ = _require_axis_info(cache_status)
            lines = build_plan_lines(
                cache_status.get("cache_id"),
                params_dict,
                axis_values,
                axis_order,
                len(cached_chunks),
                len(missing_chunks),
            )
            print("\n".join(lines))

    return _ParallelSetup(
        map_fn=map_fn,
        cached_chunks=cached_chunks,
        missing_chunks=missing_chunks,
        diagnostics=diagnostics,
        collate_fn=collate_fn,
        map_fn_kwargs=map_fn_kwargs,
    )


def _prepare_parallel_items(
    cache_status: CacheStatus,
    items: Iterable[Any],
    *,
    exec_fn: Callable[..., Any],
    cached_chunks: Sequence[ChunkKey],
    missing_chunks: Sequence[ChunkKey],
) -> tuple[
    list[Any],
    Callable[[Any], Tuple[Any, ...]] | None,
    list[list[Any]],
    list[list[Any]],
]:
    item_list = items if isinstance(items, list) else list(items)
    if not item_list:
        return item_list, None, [], []
    axis_extractor = _build_item_axis_extractor(
        cache_status,
        item_list,
        exec_fn,
    )
    cached_chunk_items = _expand_items_to_chunks_fast(
        cache_status,
        item_list,
        cached_chunks,
        axis_extractor,
    )
    missing_chunk_items = _expand_items_to_chunks_fast(
        cache_status,
        item_list,
        missing_chunks,
        axis_extractor,
    )
    return item_list, axis_extractor, cached_chunk_items, missing_chunk_items


def _scan_cached_chunk_items(
    context: RunnerContext,
    chunk_hash: ChunkHashFn,
    resolve_cache_path: ResolveCachePathFn,
    load_payload: LoadPayloadFn,
    reconstruct_output_from_items: ReconstructOutputFromItemsFn,
    reconstruct_partial_output_from_items: ReconstructPartialOutputFromItemsFn,
    build_item_maps_from_chunk_output: BuildItemMapsFromChunkOutputFn,
    write_chunk_payload: WriteChunkPayloadFn,
    item_hash: ItemHashFn,
    cached_chunks: Sequence[ChunkKey],
    cached_chunk_items: Sequence[Sequence[Any]],
    *,
    axis_extractor: Callable[[Any], Tuple[Any, ...]],
    report_progress: Callable[[int, bool], None],
    update_processed: Callable[[int], None],
    total_chunks: int,
    tracker: _MissingTracker,
    diagnostics: Diagnostics,
    cached_payloads: dict[ChunkKey, Mapping[str, Any]],
    chunk_exists: Callable[[str, Path], bool],
    load_full_chunk_payload: bool,
    collate_fn: Callable[[list[Any]], Any] | None = None,
    outputs: list[Any] | None = None,
) -> None:
    for processed, (chunk_key, chunk_items) in enumerate(
        zip(cached_chunks, cached_chunk_items), start=1
    ):
        if not chunk_items:
            report_progress(processed, processed == total_chunks)
            continue
        update_processed(len(chunk_items))
        full_chunk = len(chunk_items) == chunk_key_size(chunk_key)
        chunk_hash_value, path = resolve_chunk_path(
            chunk_hash,
            resolve_cache_path,
            chunk_key,
        )
        if not chunk_exists(chunk_hash_value, path):
            tracker.register(chunk_key, list(chunk_items))
            continue
        if full_chunk and not load_full_chunk_payload:
            diagnostics.cached_chunks += 1
            _log_chunk(context, "load", chunk_key, None)
            report_progress(processed, processed == total_chunks)
            continue
        payload = load_payload(path)
        if payload is None:
            tracker.register(chunk_key, list(chunk_items))
            continue
        if full_chunk:
            diagnostics.cached_chunks += 1
            _log_chunk(context, "load", chunk_key, None)
            if outputs is not None and collate_fn is not None:
                chunk_output = payload.get("output")
                partial = False
                if chunk_output is None:
                    items_payload = payload.get("items")
                    if items_payload is not None and isinstance(items_payload, Mapping):
                        chunk_output = reconstruct_output_from_items(
                            chunk_key, items_payload
                        )
                        if chunk_output is None:
                            chunk_output = reconstruct_partial_output_from_items(
                                chunk_key, items_payload
                            )
                            partial = True
                if chunk_output is None:
                    cached_payloads[chunk_key] = payload
                    tracker.register(chunk_key, list(chunk_items))
                    report_progress(processed, processed == total_chunks)
                    continue
                outputs.append(chunk_output)
                if partial:
                    diagnostics.partial_chunks += 1
            report_progress(processed, processed == total_chunks)
            continue
        item_map = _payload_item_map(
            build_item_maps_from_chunk_output=build_item_maps_from_chunk_output,
            resolve_cache_path=resolve_cache_path,
            write_chunk_payload=write_chunk_payload,
            chunk_key=chunk_key,
            payload=payload,
        )
        if item_map is None:
            cached_payloads[chunk_key] = payload
            tracker.register(chunk_key, list(chunk_items))
            continue
        item_outputs: list[Any] = []
        partial = False
        for item in chunk_items:
            axis_values = axis_extractor(item)
            item_key = item_hash(chunk_key, axis_values)
            if item_key not in item_map:
                partial = True
                continue
            if outputs is not None and collate_fn is not None:
                item_outputs.append(item_map[item_key])
        diagnostics.cached_chunks += 1
        _log_chunk(context, "load", chunk_key, len(chunk_items))
        if outputs is not None and collate_fn is not None and item_outputs:
            outputs.append(collate_fn(item_outputs))
        if partial:
            diagnostics.partial_chunks += 1
        report_progress(processed, processed == total_chunks)


def _exec_iter_with_progress(
    context: RunnerContext,
    exec_iter: Iterable[Any],
    *,
    cached_items_total: int,
    total_items_all: int,
    total_missing: int,
    cached_chunks: int,
    total_chunks: int,
    completed_chunk_counts: Sequence[int] | None,
) -> Iterable[tuple[int, Any]]:
    exec_start = time.monotonic()
    last_report = exec_start
    report_interval = EXEC_REPORT_INTERVAL_SECONDS
    for index, result in enumerate(exec_iter, start=1):
        now = time.monotonic()
        if context.verbose == 1 and (now - last_report) >= report_interval:
            processed_items = cached_items_total + index
            message = format_rate_eta(
                "exec_items",
                processed_items,
                total_items_all,
                exec_start,
                rate_processed=index,
                rate_total=total_missing,
            )
            completed_chunks = (
                bisect_right(completed_chunk_counts, index)
                if completed_chunk_counts is not None
                else 0
            )
            live_cached = cached_chunks + completed_chunks
            message = f"{message} cached_chunks={live_cached}/{total_chunks}"
            print_progress(message, final=False)
            last_report = now
        yield index, result
    if context.verbose == 1:
        processed_items = cached_items_total + total_missing
        message = format_rate_eta(
            "exec_items",
            processed_items,
            total_items_all,
            exec_start,
            rate_processed=total_missing,
            rate_total=total_missing,
        )
        completed_chunks = (
            bisect_right(completed_chunk_counts, total_missing)
            if completed_chunk_counts is not None
            else 0
        )
        live_cached = cached_chunks + completed_chunks
        message = f"{message} cached_chunks={live_cached}/{total_chunks}"
        print_progress(message, final=True)


def _register_missing_chunk_items(
    missing_chunks: Sequence[ChunkKey],
    missing_chunk_items: Sequence[Sequence[Any]],
    tracker: _MissingTracker,
    update_processed: Callable[[int], None],
    report_progress: Callable[[int, bool], None],
    *,
    base_index: int,
    total_chunks: int,
) -> None:
    for offset, (chunk_key, chunk_items) in enumerate(
        zip(missing_chunks, missing_chunk_items), start=1
    ):
        if chunk_items:
            tracker.register(chunk_key, list(chunk_items))
        update_processed(len(chunk_items))
        report_progress(
            base_index + offset,
            (base_index + offset) == total_chunks,
        )


def _finalize_missing_items(
    tracker: _MissingTracker,
    missing_chunks: Sequence[ChunkKey],
    item_list: Sequence[Any],
) -> list[Any]:
    if not tracker.missing_items and missing_chunks:
        return list(item_list)
    return tracker.missing_items


def _build_item_axis_extractor(
    cache_status: CacheStatus,
    items: Sequence[Any],
    exec_fn: Callable[..., Any],
) -> Callable[[Any], Tuple[Any, ...]]:
    axis_order, _ = _require_axis_info(cache_status)

    if items:
        sample = items[0]
        if isinstance(sample, Mapping):
            axis_names = list(axis_order)

            def extract(item: Any) -> Tuple[Any, ...]:
                if not isinstance(item, Mapping):
                    raise ValueError("Mixed item types: expected mapping items")
                return tuple(item[name] for name in axis_names)

            return extract
        if isinstance(sample, (tuple, list)):
            if len(sample) == len(axis_order):
                axis_count = len(axis_order)

                def extract(item: Any) -> Tuple[Any, ...]:
                    if isinstance(item, Mapping):
                        raise ValueError("Mixed item types: expected positional items")
                    if isinstance(item, tuple):
                        values = item
                    elif isinstance(item, list):
                        values = tuple(item)
                    else:
                        values = (item,)
                    if len(values) != axis_count:
                        raise ValueError(
                            "Item does not match axis order length for memo axes"
                        )
                    return tuple(values)

                return extract
        if len(axis_order) == 1:

            def extract(item: Any) -> Tuple[Any, ...]:
                if isinstance(item, Mapping):
                    raise ValueError("Mixed item types: expected positional items")
                return (item,)

            return extract

    signature = inspect.signature(exec_fn)
    param_names = list(signature.parameters)
    if param_names and param_names[0] == "params":
        axis_names = [name for name in param_names[1:] if name in axis_order]
    else:
        axis_names = [name for name in param_names if name in axis_order]
    if not axis_names:
        raise ValueError(
            "Positional items require exec_fn to list axis arguments in its signature"
        )
    axis_positions = {name: index for index, name in enumerate(axis_names)}

    def extract(item: Any) -> Tuple[Any, ...]:
        if isinstance(item, Mapping):
            raise ValueError("Mixed item types: expected positional items")
        if isinstance(item, tuple):
            values = item
        elif isinstance(item, list):
            values = tuple(item)
        else:
            values = (item,)
        if len(values) < len(axis_names):
            raise ValueError("Item does not provide enough positional values")
        values_by_axis = {
            axis_name: values[index] for axis_name, index in axis_positions.items()
        }
        missing = [name for name in axis_order if name not in values_by_axis]
        if missing:
            raise ValueError(
                f"Item does not provide positional values for axes: {missing}"
            )
        return tuple(values_by_axis[name] for name in axis_order)

    return extract


def _expand_items_to_chunks_fast(
    cache_status: CacheStatus,
    items: Sequence[Any],
    chunk_keys: Sequence[ChunkKey],
    axis_extractor: Callable[[Any], Tuple[Any, ...]],
) -> list[list[Any]]:
    if not chunk_keys:
        return [[] for _ in chunk_keys]
    axis_order, axis_chunk_maps = _require_axis_info(cache_status)
    chunk_index_map: dict[Tuple[int, ...], int] = {}
    for index, chunk_key in enumerate(chunk_keys):
        chunk_ids: list[int] = []
        for axis, values in chunk_key:
            chunk_ids.append(axis_chunk_maps[axis][values[0]])
        chunk_index_map[tuple(chunk_ids)] = index

    chunked: list[list[Any]] = [[] for _ in chunk_keys]
    for item in items:
        axis_vals = axis_extractor(item)
        if len(axis_vals) != len(axis_order):
            raise ValueError("Item axis values do not match axis order")
        chunk_ids = []
        for axis, value in zip(axis_order, axis_vals):
            if value not in axis_chunk_maps[axis]:
                raise KeyError(
                    f"Value '{value}' not found in axis_values for axis '{axis}'"
                )
            chunk_ids.append(axis_chunk_maps[axis][value])
        chunk_index = chunk_index_map.get(tuple(chunk_ids))
        if chunk_index is None:
            continue
        chunked[chunk_index].append(item)
    return chunked


def _extract_axis_values(
    items: Sequence[Any], axis_extractor: Callable[[Any], Tuple[Any, ...]]
) -> list[Tuple[Any, ...]]:
    return [axis_extractor(item) for item in items]


def _collect_missing_axis_values(
    *,
    items: Sequence[Any],
    axis_order: Sequence[str],
    axis_values: Mapping[str, Sequence[Any]] | None,
    axis_extractor: Callable[[Any], Tuple[Any, ...]],
) -> dict[str, list[Any]]:
    if axis_values is None:
        return {}
    existing_sets = {axis: set(axis_values.get(axis, [])) for axis in axis_order}
    missing: dict[str, list[Any]] = {}
    for item in items:
        axis_vals = axis_extractor(item)
        for axis, value in zip(axis_order, axis_vals):
            if value in existing_sets[axis]:
                continue
            missing.setdefault(axis, []).append(value)
            existing_sets[axis].add(value)
    return missing


def _collect_axis_values_from_items(
    *,
    items: Sequence[Any],
    axis_order: Sequence[str],
    axis_values: Mapping[str, Sequence[Any]] | None,
    axis_extractor: Callable[[Any], Tuple[Any, ...]],
) -> dict[str, list[Any]]:
    collected: dict[str, list[Any]] = {axis: [] for axis in axis_order}
    seen: dict[str, set[Any]] = {axis: set() for axis in axis_order}
    for item in items:
        axis_vals = axis_extractor(item)
        for axis, value in zip(axis_order, axis_vals):
            if value in seen[axis]:
                continue
            seen[axis].add(value)
            collected[axis].append(value)
    if axis_values is not None:
        ordered: dict[str, list[Any]] = {}
        for axis in axis_order:
            if axis not in seen:
                continue
            full_values = axis_values.get(axis, [])
            ordered_values = [value for value in full_values if value in seen[axis]]
            if ordered_values:
                ordered[axis] = ordered_values
        return ordered
    return {axis: values for axis, values in collected.items() if values}


def _cumulative_chunk_counts(
    chunk_order: Sequence[ChunkKey],
    items_by_chunk: Mapping[ChunkKey, Sequence[Any]],
) -> list[int]:
    counts: list[int] = []
    running = 0
    for chunk_key in chunk_order:
        chunk_size = len(items_by_chunk.get(chunk_key, []))
        if chunk_size <= 0:
            continue
        running += chunk_size
        counts.append(running)
    return counts


def run_parallel(
    items: Iterable[Any],
    *,
    exec_fn: Callable[..., Any],
    cache: CacheProtocol,
    map_fn: Callable[..., Iterable[Any]] | None = None,
    map_fn_kwargs: Mapping[str, Any] | None = None,
    collate_fn: Callable[[list[Any]], Any] | None = None,
    flush_on_chunk: bool = False,
    return_output: bool = True,
    extend_cache: bool = False,
    params: dict[str, Any] | None = None,
    axis_values_override: Mapping[str, Sequence[Any]] | None = None,
    allow_superset: bool = False,
    # Manual cache methods (optional overrides)
    write_metadata: WriteMetadataFn | None = None,
    chunk_hash: ChunkHashFn | None = None,
    resolve_cache_path: ResolveCachePathFn | None = None,
    load_payload: LoadPayloadFn | None = None,
    write_chunk_payload: WriteChunkPayloadFn | None = None,
    update_chunk_index: UpdateChunkIndexFn | None = None,
    load_chunk_index: LoadChunkIndexFn | None = None,
    build_item_maps_from_axis_values: BuildItemMapsFromAxisValuesFn | None = None,
    build_item_maps_from_chunk_output: BuildItemMapsFromChunkOutputFn | None = None,
    reconstruct_output_from_items: ReconstructOutputFromItemsFn | None = None,
    collect_chunk_data: CollectChunkDataFn | None = None,
    item_hash: ItemHashFn | None = None,
    context: RunnerContext | None = None,
) -> tuple[Any, Diagnostics]:
    """Execute items in parallel, reusing cached chunk data.

    cache must already represent the desired axis subset.
    collate_fn overrides cache.collate_fn for this run.
    Partial chunks return whatever cached output is available and increment
    diagnostics.partial_chunks.
    flush_on_chunk writes chunk payloads as soon as a chunk completes.
    If return_output is False, chunk payloads are flushed incrementally and
    the output is returned as None.
    extend_cache updates the cache axis_values in-place if items introduce
    new axis values.
    """
    cache = resolve_cache_for_run(
        cache,
        params=params,
        axis_values_override=axis_values_override,
        extend_cache=extend_cache,
        allow_superset=allow_superset,
    )
    cache_status = cache.cache_status()
    if not return_output:
        flush_on_chunk = True
    deps = resolve_runner_deps(
        cache=cache,
        context=context,
        write_metadata=write_metadata,
        chunk_hash=chunk_hash,
        resolve_cache_path=resolve_cache_path,
        load_payload=load_payload,
        write_chunk_payload=write_chunk_payload,
        update_chunk_index=update_chunk_index,
        load_chunk_index=load_chunk_index,
        build_item_maps_from_axis_values=build_item_maps_from_axis_values,
        build_item_maps_from_chunk_output=build_item_maps_from_chunk_output,
        reconstruct_output_from_items=reconstruct_output_from_items,
        collect_chunk_data=collect_chunk_data,
        item_hash=item_hash,
        require=[
            "context",
            "write_metadata",
            "chunk_hash",
            "resolve_cache_path",
            "load_payload",
            "write_chunk_payload",
            "update_chunk_index",
            "build_item_maps_from_axis_values",
            "build_item_maps_from_chunk_output",
            "reconstruct_output_from_items",
            "reconstruct_partial_output_from_items",
            "item_hash",
        ],
    )
    context = cast(RunnerContext, deps.context)
    item_list = items if isinstance(items, list) else list(items)
    axis_extractor = _build_item_axis_extractor(cache_status, item_list, exec_fn)
    if extend_cache and axis_extractor is not None:
        axis_order, _ = _require_axis_info(cache_status)
        missing_axis_values = _collect_missing_axis_values(
            items=item_list,
            axis_order=axis_order,
            axis_values=cache_status.get("axis_values"),
            axis_extractor=axis_extractor,
        )
        if missing_axis_values:
            cache.extend_axis_values(missing_axis_values)
            cache_status = cache.cache_status()

    cached_chunks_all = list(cache_status.get("cached_chunks", []))
    missing_chunks_all = list(cache_status.get("missing_chunks", []))
    cached_chunk_indices_all = list(cache_status.get("cached_chunk_indices", []))
    missing_chunk_indices_all = list(cache_status.get("missing_chunk_indices", []))

    requested_axes: dict[str, list[Any]] = {}
    if axis_extractor is not None:
        axis_order, _ = _require_axis_info(cache_status)
        requested_axes = _collect_axis_values_from_items(
            items=item_list,
            axis_order=axis_order,
            axis_values=cache_status.get("axis_values"),
            axis_extractor=axis_extractor,
        )

    cached_chunk_items_all = _expand_items_to_chunks_fast(
        cache_status,
        item_list,
        cached_chunks_all,
        axis_extractor,
    )
    missing_chunk_items_all = _expand_items_to_chunks_fast(
        cache_status,
        item_list,
        missing_chunks_all,
        axis_extractor,
    )
    cached_mask = [bool(items) for items in cached_chunk_items_all]
    missing_mask = [bool(items) for items in missing_chunk_items_all]

    cached_chunks = [
        chunk for chunk, keep in zip(cached_chunks_all, cached_mask) if keep
    ]
    missing_chunks = [
        chunk for chunk, keep in zip(missing_chunks_all, missing_mask) if keep
    ]
    cached_chunk_indices = [
        index for index, keep in zip(cached_chunk_indices_all, cached_mask) if keep
    ]
    missing_chunk_indices = [
        index for index, keep in zip(missing_chunk_indices_all, missing_mask) if keep
    ]
    cached_chunk_items = [
        items for items, keep in zip(cached_chunk_items_all, cached_mask) if keep
    ]
    missing_chunk_items = [
        items for items, keep in zip(missing_chunk_items_all, missing_mask) if keep
    ]

    cache_status_for_plan = dict(cache_status)
    cache_status_for_plan["cached_chunks"] = cached_chunks
    cache_status_for_plan["missing_chunks"] = missing_chunks
    cache_status_for_plan["cached_chunk_indices"] = cached_chunk_indices
    cache_status_for_plan["missing_chunk_indices"] = missing_chunk_indices

    setup = _prepare_parallel_setup(
        cast(RunnerContext, deps.context),
        cast(WriteMetadataFn, deps.write_metadata),
        cast(CacheStatus, cache_status_for_plan),
        map_fn=map_fn,
        map_fn_kwargs=map_fn_kwargs,
        collate_fn=collate_fn,
        axis_values_override=requested_axes or None,
    )
    diagnostics = setup.diagnostics
    collate_fn_resolved = setup.collate_fn
    map_fn_resolved = setup.map_fn
    map_fn_kwargs_resolved = setup.map_fn_kwargs
    outputs: list[Any] | None = [] if return_output else None
    exec_outputs: list[Any] = []
    total_chunks = diagnostics.total_chunks
    (
        item_list,
        axis_extractor,
        cached_chunk_items,
        missing_chunk_items,
    ) = _prepare_parallel_items(
        cache_status,
        item_list,
        exec_fn=exec_fn,
        cached_chunks=cached_chunks,
        missing_chunks=missing_chunks,
    )
    if not item_list or axis_extractor is None:
        return [], diagnostics
    total_items = len(item_list)
    report_progress_main, update_processed, set_allow_final = prepare_planning_progress(
        total_chunks=total_chunks,
        total_items=total_items,
        verbose=context.verbose,
        label="planning",
    )

    tracker = _MissingTracker.create(track_item_keys=True)
    cached_payloads: dict[ChunkKey, Mapping[str, Any]] = {}

    use_index = False
    chunk_index: dict[str, Any] = {}
    load_chunk_index_fn = cast(LoadChunkIndexFn | None, deps.load_chunk_index)
    if load_chunk_index_fn is not None:
        chunk_index = load_chunk_index_fn() or {}
        use_index = bool(chunk_index)

    _scan_cached_chunk_items(
        context,
        cast(ChunkHashFn, deps.chunk_hash),
        cast(ResolveCachePathFn, deps.resolve_cache_path),
        cast(LoadPayloadFn, deps.load_payload),
        cast(ReconstructOutputFromItemsFn, deps.reconstruct_output_from_items),
        cast(
            ReconstructPartialOutputFromItemsFn,
            deps.reconstruct_partial_output_from_items,
        ),
        cast(BuildItemMapsFromChunkOutputFn, deps.build_item_maps_from_chunk_output),
        cast(WriteChunkPayloadFn, deps.write_chunk_payload),
        cast(ItemHashFn, deps.item_hash),
        cached_chunks,
        cached_chunk_items,
        axis_extractor=axis_extractor,
        report_progress=report_progress_main,
        update_processed=update_processed,
        total_chunks=total_chunks,
        tracker=tracker,
        diagnostics=diagnostics,
        cached_payloads=cached_payloads,
        chunk_exists=lambda _chunk_hash, path: (
            _chunk_hash in chunk_index if use_index else path.exists()
        ),
        load_full_chunk_payload=return_output,
        collate_fn=collate_fn_resolved,
        outputs=outputs,
    )

    base_index = len(cached_chunks)
    _register_missing_chunk_items(
        missing_chunks,
        missing_chunk_items,
        tracker,
        update_processed,
        report_progress_main,
        base_index=base_index,
        total_chunks=total_chunks,
    )

    missing_items = _finalize_missing_items(tracker, missing_chunks, item_list)
    if not missing_items and context.verbose == 1:
        set_allow_final(True)
        report_progress_main(total_chunks, True)

    if missing_items:
        diagnostics.executed_chunks = len(tracker.missing_chunk_order)
        exec_fn = functools.partial(_exec_with_item, exec_fn)
        exec_iter = map_fn_resolved(
            exec_fn,
            missing_items,
            **map_fn_kwargs_resolved,
        )
        total_items_all = len(item_list)
        cached_items_total = total_items_all - len(missing_items)
        completed_chunk_counts = _cumulative_chunk_counts(
            tracker.missing_chunk_order,
            tracker.missing_items_by_chunk,
        )
        missing_item_keys = tracker.missing_item_keys or []
        queued_chunks: dict[ChunkKey, list[Any]] = {}
        queued_outputs: dict[ChunkKey, list[Any]] = {}

        for item_idx, result in _exec_iter_with_progress(
            context,
            exec_iter,
            cached_items_total=cached_items_total,
            total_items_all=total_items_all,
            total_missing=len(missing_items),
            cached_chunks=diagnostics.cached_chunks,
            total_chunks=total_chunks,
            completed_chunk_counts=completed_chunk_counts,
        ):
            chunk_key = missing_item_keys[item_idx - 1]
            queued_chunks.setdefault(chunk_key, []).append(missing_items[item_idx - 1])
            queued_outputs.setdefault(chunk_key, []).append(result)
            if not flush_on_chunk:
                exec_outputs.append(result)
                continue
            if len(queued_outputs[chunk_key]) < chunk_key_size(chunk_key):
                continue
            chunk_items = queued_chunks.pop(chunk_key, [])
            chunk_outputs = queued_outputs.pop(chunk_key, [])
            chunk_output = collate_fn_resolved(chunk_outputs)
            axis_values = _extract_axis_values(chunk_items, axis_extractor)
            item_map, item_axis_vals = cast(
                BuildItemMapsFromAxisValuesFn, deps.build_item_maps_from_axis_values
            )(
                chunk_key,
                axis_values,
                chunk_outputs,
            )
            chunk_hash_value = cast(ChunkHashFn, deps.chunk_hash)(chunk_key)
            _save_chunk_payload(
                resolve_cache_path=cast(ResolveCachePathFn, deps.resolve_cache_path),
                write_chunk_payload=cast(WriteChunkPayloadFn, deps.write_chunk_payload),
                update_chunk_index=cast(UpdateChunkIndexFn, deps.update_chunk_index),
                chunk_key=chunk_key,
                chunk_output=chunk_output,
                item_map=item_map,
                cached_payloads=cached_payloads,
                chunk_hash=chunk_hash_value,
                missing_chunks=missing_chunks,
                spec_fn=lambda: item_axis_vals,
            )
            _log_chunk(context, "run", chunk_key, len(chunk_items))
            if outputs is not None:
                outputs.append(chunk_output)
            diagnostics.max_parallel_items = max(
                diagnostics.max_parallel_items, len(chunk_items)
            )
            diagnostics.stream_flushes += 1

        if flush_on_chunk and queued_outputs:
            for chunk_key, chunk_outputs in queued_outputs.items():
                chunk_items = queued_chunks.get(chunk_key, [])
                if not chunk_items:
                    continue
                chunk_output = collate_fn_resolved(chunk_outputs)
                axis_values = _extract_axis_values(chunk_items, axis_extractor)
                item_map, item_axis_vals = cast(
                    BuildItemMapsFromAxisValuesFn, deps.build_item_maps_from_axis_values
                )(
                    chunk_key,
                    axis_values,
                    chunk_outputs,
                )
                chunk_hash_value = cast(ChunkHashFn, deps.chunk_hash)(chunk_key)
                _save_chunk_payload(
                    resolve_cache_path=cast(
                        ResolveCachePathFn, deps.resolve_cache_path
                    ),
                    write_chunk_payload=cast(
                        WriteChunkPayloadFn, deps.write_chunk_payload
                    ),
                    update_chunk_index=cast(
                        UpdateChunkIndexFn, deps.update_chunk_index
                    ),
                    chunk_key=chunk_key,
                    chunk_output=chunk_output,
                    item_map=item_map,
                    cached_payloads=cached_payloads,
                    chunk_hash=chunk_hash_value,
                    missing_chunks=missing_chunks,
                    spec_fn=lambda: item_axis_vals,
                )
                _log_chunk(context, "run", chunk_key, len(chunk_items))
                if outputs is not None:
                    outputs.append(chunk_output)
                diagnostics.max_parallel_items = max(
                    diagnostics.max_parallel_items, len(chunk_items)
                )
                diagnostics.stream_flushes += 1
                if len(chunk_items) < chunk_key_size(chunk_key):
                    diagnostics.partial_chunks += 1

        if not flush_on_chunk:
            cursor = 0
            processed_missing_chunks = 0
            for chunk_key in tracker.missing_chunk_order:
                chunk_items = tracker.missing_items_by_chunk.get(chunk_key, [])
                chunk_size = len(chunk_items)
                if chunk_size == 0:
                    continue
                chunk_outputs = exec_outputs[cursor : cursor + chunk_size]
                chunk_output = collate_fn_resolved(chunk_outputs)
                axis_values = _extract_axis_values(chunk_items, axis_extractor)
                item_map, item_axis_vals = cast(
                    BuildItemMapsFromAxisValuesFn, deps.build_item_maps_from_axis_values
                )(
                    chunk_key,
                    axis_values,
                    chunk_outputs,
                )
                chunk_hash_value = cast(ChunkHashFn, deps.chunk_hash)(chunk_key)
                _save_chunk_payload(
                    resolve_cache_path=cast(
                        ResolveCachePathFn, deps.resolve_cache_path
                    ),
                    write_chunk_payload=cast(
                        WriteChunkPayloadFn, deps.write_chunk_payload
                    ),
                    update_chunk_index=cast(
                        UpdateChunkIndexFn, deps.update_chunk_index
                    ),
                    chunk_key=chunk_key,
                    chunk_output=chunk_output,
                    item_map=item_map,
                    cached_payloads=cached_payloads,
                    chunk_hash=chunk_hash_value,
                    missing_chunks=missing_chunks,
                    spec_fn=lambda: item_axis_vals,
                )
                _log_chunk(context, "run", chunk_key, chunk_size)
                if outputs is not None:
                    outputs.append(chunk_output)
                cursor += chunk_size
                processed_missing_chunks += 1
                report_progress_main(base_index + processed_missing_chunks, False)

    merged: Any = None
    if outputs is not None:
        merged = _merge_outputs(
            context,
            outputs,
            diagnostics,
            collate_fn=collate_fn_resolved,
        )
        if not merged and item_list:
            merged = exec_outputs if missing_items else []
    print_chunk_summary(diagnostics, context.verbose)
    return merged, diagnostics



"""Runner protocols and cache status shape."""

from __future__ import annotations

from pathlib import Path
from typing import Any, Callable, Mapping, Protocol, Sequence, Tuple, TypedDict

ChunkKey = Tuple[Tuple[str, Tuple[Any, ...]], ...]
AxisChunkMaps = dict[str, dict[Any, int]]


class CacheStatus(TypedDict):
    """Cache status payload consumed by runners."""

    cache_id: str
    metadata: dict[str, Any]
    axis_values: Mapping[str, Any]
    axis_order: Tuple[str, ...]
    axis_chunk_maps: AxisChunkMaps
    cached_chunks: list[ChunkKey]
    missing_chunks: list[ChunkKey]

    total_chunks: int
    cached_chunk_indices: list[dict[str, Any]]
    missing_chunk_indices: list[dict[str, Any]]


class RunnerContext(Protocol):
    """Runtime settings consumed by runner helpers."""

    verbose: int
    profile: bool
    collate_fn: Callable[[list[Any]], Any] | None


WriteMetadataFn = Callable[[], Path]
ChunkHashFn = Callable[[ChunkKey], str]
ResolveCachePathFn = Callable[[ChunkKey, str], Path]
LoadPayloadFn = Callable[[Path], dict[str, Any] | None]
WriteChunkPayloadFn = Callable[..., Path]
UpdateChunkIndexFn = Callable[[str, ChunkKey], None]
LoadChunkIndexFn = Callable[[], dict[str, Any] | None]
BuildItemMapsFromAxisValuesFn = Callable[
    ...,
    tuple[dict[str, Any], dict[str, dict[str, Any]]],
]
BuildItemMapsFromChunkOutputFn = Callable[
    ...,
    tuple[dict[str, Any] | None, dict[str, dict[str, Any]] | None],
]
ExtractItemsFromMapFn = Callable[..., list[Any] | None]
ReconstructOutputFromItemsFn = Callable[[ChunkKey, Mapping[str, Any]], list[Any] | None]
ReconstructPartialOutputFromItemsFn = Callable[[ChunkKey, Mapping[str, Any]], list[Any]]
CollectChunkDataFn = Callable[..., tuple[Any | None, bool]]
ItemHashFn = Callable[[ChunkKey, Tuple[Any, ...]], str]


class CacheProtocol(RunnerContext, Protocol):
    """Protocol for cache objects used by runners."""

    def cache_status(
        self,
        *,
        axis_indices: Mapping[str, Any] | None = None,
        axis_values_override: Mapping[str, Sequence[Any]] | None = None,
        extend_cache: bool = False,
        **axes: Any,
    ) -> CacheStatus: ...

    def write_metadata(self) -> Path: ...

    def chunk_hash(self, chunk_key: ChunkKey) -> str: ...

    def resolve_cache_path(self, chunk_key: ChunkKey, chunk_hash: str) -> Path: ...

    def load_payload(self, path: Path) -> dict[str, Any] | None: ...

    def write_chunk_payload(self, *args: Any, **kwargs: Any) -> Path: ...

    def update_chunk_index(self, chunk_hash: str, chunk_key: ChunkKey) -> None: ...

    def build_item_maps_from_axis_values(self, *args: Any, **kwargs: Any) -> Any: ...

    def build_item_maps_from_chunk_output(self, *args: Any, **kwargs: Any) -> Any: ...

    def reconstruct_output_from_items(
        self, chunk_key: ChunkKey, items: Mapping[str, Any]
    ) -> list[Any] | None: ...

    def reconstruct_partial_output_from_items(
        self, chunk_key: ChunkKey, items: Mapping[str, Any]
    ) -> list[Any]: ...

    def collect_chunk_data(
        self, *args: Any, **kwargs: Any
    ) -> tuple[Any | None, bool]: ...

    def extend_axis_values(
        self, axis_values: Mapping[str, Sequence[Any]], *, write_metadata: bool = True
    ) -> None: ...

    def extract_items_from_map(self, *args: Any, **kwargs: Any) -> list[Any] | None: ...

    def item_hash(self, chunk_key: ChunkKey, axis_values: Tuple[Any, ...]) -> str: ...

    def load_chunk_index(self) -> dict[str, Any] | None: ...

    def requested_items_by_chunk(
        self,
    ) -> Mapping[ChunkKey, list[Tuple[Any, ...]]] | None: ...



import json
from pathlib import Path
from typing import Any, Tuple

from .data_write_utils import _apply_payload_timestamps, _atomic_write_json

ChunkKey = Tuple[Tuple[str, Tuple[Any, ...]], ...]


def chunk_index_path(memo_root: Path) -> Path:
    return memo_root / "chunks_index.json"


def load_chunk_index(memo_root: Path) -> dict[str, Any]:
    path = chunk_index_path(memo_root)
    if not path.exists():
        return {}
    try:
        with open(path, "r", encoding="utf-8") as handle:
            data = json.load(handle)
    except json.JSONDecodeError:
        return {}
    if isinstance(data, dict):
        return data
    return {}


def update_chunk_index(
    memo_root: Path,
    chunk_hash: str,
    chunk_key: ChunkKey,
) -> None:
    index = load_chunk_index(memo_root)
    existing_entry = index.get(chunk_hash)
    entry = {
        "chunk_key": chunk_key,
    }
    _apply_payload_timestamps(entry, existing=existing_entry)
    index[chunk_hash] = entry
    _atomic_write_json(chunk_index_path(memo_root), index)



"""Memo wrapper helpers and re-exports."""

from __future__ import annotations

import functools
import inspect
import itertools
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Any, Callable, Iterable, Mapping, Sequence, Tuple, cast

from .cache import CachePathFn, ChunkCache, CollateFn, MemoChunkEnumerator
from .identity import params_to_cache_id, stable_serialize
from .runners import Diagnostics, run, run_parallel, run_streaming
from .runners_common import resolve_cache_for_run


class ChunkMemo:
    """Chunk cache manager with decorator helpers."""

    def __init__(
        self,
        root: str | Path | None,
        chunk_spec: dict[str, Any] | None,
        axis_values: dict[str, Any],
        *,
        metadata: dict[str, Any] | None = None,
        collate_fn: CollateFn | None = None,
        chunk_enumerator: MemoChunkEnumerator | None = None,
        chunk_hash_fn: Callable[
            [str, Tuple[Tuple[str, Tuple[Any, ...]], ...], str], str
        ]
        | None = None,
        path_fn: CachePathFn | None = None,
        version: str = "v1",
        axis_order: Sequence[str] | None = None,
        verbose: int = 1,
        profile: bool = False,
        exclusive: bool = False,
        warn_on_overlap: bool = False,
        precompute_chunk_keys: bool = False,
    ) -> None:
        self.root = root
        self.chunk_spec = chunk_spec
        self.axis_values = axis_values
        self.metadata = metadata or {}
        self.collate_fn = collate_fn
        self.chunk_enumerator = chunk_enumerator
        self.chunk_hash_fn = chunk_hash_fn
        self.path_fn = path_fn
        self.version = version
        self.axis_order = axis_order
        self.verbose = verbose
        self.profile = profile
        self.exclusive = exclusive
        self.warn_on_overlap = warn_on_overlap
        self.precompute_chunk_keys = precompute_chunk_keys
        self._caches: dict[str, ChunkCache] = {}

    @classmethod
    def auto_load(
        cls,
        root: str | Path,
        params: dict[str, Any],
        axis_values: dict[str, Any] | None = None,
        chunk_spec: dict[str, Any] | None = None,
        metadata: dict[str, Any] | None = None,
        collate_fn: CollateFn | None = None,
        chunk_enumerator: MemoChunkEnumerator | None = None,
        chunk_hash_fn: Callable[
            [str, Tuple[Tuple[str, Tuple[Any, ...]], ...], str], str
        ]
        | None = None,
        path_fn: CachePathFn | None = None,
        version: str = "v1",
        axis_order: Sequence[str] | None = None,
        verbose: int = 1,
        profile: bool = False,
        exclusive: bool = False,
        warn_on_overlap: bool = False,
        precompute_chunk_keys: bool = False,
        allow_superset: bool = False,
    ) -> "ChunkMemo":
        if not isinstance(params, dict):
            raise ValueError("'params' must be a dict")
        cache_id = params_to_cache_id(params)
        root_path = Path(root)
        axis_values_map = axis_values or {}

        def summarize_axis_values(value: Any) -> str:
            if not isinstance(value, Mapping):
                return repr(value)
            summarized: dict[str, Any] = {}
            for key, axis_vals in value.items():
                if isinstance(axis_vals, (list, tuple)):
                    if len(axis_vals) <= 6:
                        summarized[key] = list(axis_vals)
                    else:
                        summarized[key] = (
                            list(axis_vals[:3]) + ["..."] + list(axis_vals[-2:])
                        )
                else:
                    summarized[key] = axis_vals
            rendered = repr(summarized)
            if len(rendered) > 200:
                return rendered[:197] + "..."
            return rendered

        cache_path = root_path / cache_id
        base_metadata = dict(metadata or {})
        base_metadata.pop("params", None)

        if cache_path.exists():
            cache = ChunkCache.load_from_cache(
                root=root_path,
                cache_id=cache_id,
                collate_fn=collate_fn,
                chunk_enumerator=chunk_enumerator,
                chunk_hash_fn=chunk_hash_fn,
                path_fn=path_fn,
                verbose=verbose,
                profile=profile,
                exclusive=exclusive,
                warn_on_overlap=warn_on_overlap,
            )
            if not base_metadata:
                base_metadata = dict(cache.metadata)
                base_metadata.pop("params", None)
            if axis_values is not None:
                meta_axis_values = cache._axis_values_serializable
                if meta_axis_values is None:
                    raise ValueError("Cache missing axis_values")
                if allow_superset:
                    if not ChunkCache._is_superset_compatible(
                        {"axis_values": meta_axis_values}, axis_values
                    ):
                        raise ValueError(
                            "Requested axis_values are not a subset of existing cache. "
                            f"Existing axis_values={summarize_axis_values(meta_axis_values)}, "
                            f"requested={summarize_axis_values(axis_values)}."
                        )
                    cache = cache.slice(**axis_values)
                elif stable_serialize(meta_axis_values) != stable_serialize(
                    axis_values
                ):
                    raise ValueError(
                        "Cache with same cache_id exists but axis_values differ. "
                        f"Existing axis_values={summarize_axis_values(meta_axis_values)}, "
                        f"requested={summarize_axis_values(axis_values)}."
                    )
            merged_metadata = dict(base_metadata)
            merged_metadata["params"] = params
            if cache.metadata != merged_metadata:
                cache.set_identity(cache.cache_id, metadata=merged_metadata)
            memo = cls(
                root=root_path,
                chunk_spec=cache.chunk_spec,
                axis_values=(
                    axis_values
                    if axis_values is not None
                    else (cache._axis_values_serializable or {})
                ),
                metadata=base_metadata,
                collate_fn=cache.collate_fn,
                chunk_enumerator=cache.chunk_enumerator,
                chunk_hash_fn=cache.chunk_hash_fn,
                path_fn=cache.path_fn,
                version=cache.version,
                axis_order=cache.axis_order,
                verbose=cache.verbose,
                profile=cache.profile,
                exclusive=cache.exclusive,
                warn_on_overlap=cache.warn_on_overlap,
                precompute_chunk_keys=cache.precompute_chunk_keys,
            )
            memo._caches[cache_id] = cache
            return memo

        if axis_values is None:
            raise ValueError("axis_values is required when creating a new cache")
        merged_metadata = dict(base_metadata)
        merged_metadata["params"] = params
        cache = ChunkCache(
            root=root_path,
            cache_id=cache_id,
            metadata=merged_metadata,
            chunk_spec=chunk_spec,
            axis_values=axis_values_map,
            collate_fn=collate_fn,
            chunk_enumerator=chunk_enumerator,
            chunk_hash_fn=chunk_hash_fn,
            path_fn=path_fn,
            version=version,
            axis_order=axis_order,
            verbose=verbose,
            profile=profile,
            exclusive=exclusive,
            warn_on_overlap=warn_on_overlap,
            precompute_chunk_keys=precompute_chunk_keys,
        )
        memo = cls(
            root=root_path,
            chunk_spec=cache.chunk_spec,
            axis_values=axis_values_map,
            metadata=base_metadata,
            collate_fn=collate_fn,
            chunk_enumerator=chunk_enumerator,
            chunk_hash_fn=chunk_hash_fn,
            path_fn=path_fn,
            version=version,
            axis_order=axis_order,
            verbose=verbose,
            profile=profile,
            exclusive=exclusive,
            warn_on_overlap=warn_on_overlap,
            precompute_chunk_keys=precompute_chunk_keys,
        )
        memo._caches[cache_id] = cache
        return memo

    def cache_for_params(self, params: dict[str, Any]) -> ChunkCache:
        if not isinstance(params, dict):
            raise ValueError("'params' must be a dict")
        cache_id = params_to_cache_id(params)
        cache = self._caches.get(cache_id)
        merged_metadata = dict(self.metadata)
        merged_metadata["params"] = params
        if cache is None:
            cache = ChunkCache(
                root=self.root,
                cache_id=cache_id,
                metadata=merged_metadata,
                chunk_spec=self.chunk_spec,
                axis_values=self.axis_values,
                collate_fn=self.collate_fn,
                chunk_enumerator=self.chunk_enumerator,
                chunk_hash_fn=self.chunk_hash_fn,
                path_fn=self.path_fn,
                version=self.version,
                axis_order=self.axis_order,
                verbose=self.verbose,
                profile=self.profile,
                exclusive=self.exclusive,
                warn_on_overlap=self.warn_on_overlap,
                precompute_chunk_keys=self.precompute_chunk_keys,
            )
            self._caches[cache_id] = cache
        elif cache.metadata != merged_metadata:
            cache.set_identity(cache_id, metadata=merged_metadata)
        return cache

    def cache(
        self,
        *,
        params_arg: str = "params",
        map_fn: Callable[..., Iterable[Any]] | None = None,
        map_fn_kwargs: Mapping[str, Any] | None = None,
        max_workers: int = 1,
    ) -> Callable[[Callable[..., Any]], Callable[..., Tuple[Any, Diagnostics]]]:
        """Decorator for running memoized execution with output.

        Infers memoization params from non-axis arguments unless an explicit
        params dict is provided. Supports axis selection by value or by index
        via axis_indices. Uses runners.run under the hood.
        """
        return self._build_wrapper(
            params_arg=params_arg,
            streaming=False,
            map_fn=map_fn,
            map_fn_kwargs=map_fn_kwargs,
            max_workers=max_workers,
        )

    def stream_cache(
        self, *, params_arg: str = "params"
    ) -> Callable[[Callable[..., Any]], Callable[..., Diagnostics]]:
        """Decorator for streaming memoized execution to disk only.

        Infers memoization params from non-axis arguments unless an explicit
        params dict is provided. Supports axis selection by value or by index
        via axis_indices. Uses runners.run_streaming under the hood.
        """
        return self._build_wrapper(params_arg=params_arg, streaming=True)

    def _split_bound_args(
        self,
        bound_args: Mapping[str, Any],
        params_arg: str,
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        axis_names = set(self.axis_values or {})
        extras = {k: v for k, v in bound_args.items() if k != params_arg}
        axis_inputs = {k: v for k, v in extras.items() if k in axis_names}
        exec_extras = {k: v for k, v in extras.items() if k not in axis_names}
        return exec_extras, axis_inputs

    def _merge_params(
        self,
        base_params: dict[str, Any] | None,
        exec_extras: dict[str, Any],
        *,
        params_provided: bool,
    ) -> dict[str, Any]:
        merged_params = dict(base_params or {})
        if params_provided:
            merged_params = dict(exec_extras)
            merged_params.update(base_params or {})
        else:
            merged_params.update(exec_extras)
        return merged_params

    def _build_wrapper(
        self,
        *,
        params_arg: str,
        streaming: bool,
        map_fn: Callable[..., Iterable[Any]] | None = None,
        map_fn_kwargs: Mapping[str, Any] | None = None,
        max_workers: int = 1,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
            signature = inspect.signature(func)

            @functools.wraps(func)
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                if params_arg in kwargs and args:
                    raise ValueError(
                        f"'{params_arg}' passed as both positional and keyword"
                    )

                axis_indices = kwargs.pop("axis_indices", None)
                extend_cache = kwargs.pop("extend_cache", False)
                bound = signature.bind_partial(*args, **kwargs)
                bound.apply_defaults()
                params_provided = params_arg in bound.arguments
                params = bound.arguments.get(params_arg)
                if params_provided and not isinstance(params, dict):
                    raise ValueError(f"'{params_arg}' must be a dict")

                exec_extras, axis_inputs = self._split_bound_args(
                    bound.arguments, params_arg
                )
                base_params: dict[str, Any] | None = (
                    params if params_provided else self.metadata.get("params", {})
                )
                merged_params = self._merge_params(
                    base_params,
                    exec_extras,
                    params_provided=params_provided,
                )
                cache = resolve_cache_for_run(
                    self,
                    params=merged_params,
                    axis_values_override=axis_inputs if extend_cache else None,
                    extend_cache=extend_cache,
                    allow_superset=extend_cache,
                )

                exec_kwargs = dict(exec_extras)
                if params_arg in signature.parameters:
                    exec_kwargs[params_arg] = merged_params
                exec_fn = functools.partial(func, **exec_kwargs)
                sliced = cast(ChunkCache, cache).slice(
                    axis_indices=axis_indices, **axis_inputs
                )
                if streaming:
                    return run_streaming(sliced, exec_fn)
                if max_workers > 1 or map_fn is not None:
                    if map_fn is None:
                        _require_top_level_function(func)
                    return run_parallel(
                        _build_items_for_parallel(sliced),
                        exec_fn=exec_fn,
                        cache=sliced,
                        map_fn=map_fn or _map_process_pool,
                        map_fn_kwargs=_resolve_map_kwargs(map_fn_kwargs, max_workers),
                    )
                return run(sliced, exec_fn)

            def cache_status(
                params: dict[str, Any] | None = None,
                *,
                axis_indices: Mapping[str, Any] | None = None,
                **axes: Any,
            ):
                if params is not None and not isinstance(params, dict):
                    raise ValueError(f"'{params_arg}' must be a dict")
                exec_extras, axis_inputs = self._split_bound_args(axes, params_arg)
                base_params = (
                    params if params is not None else self.metadata.get("params", {})
                )
                merged_params = self._merge_params(
                    base_params,
                    exec_extras,
                    params_provided=params is not None,
                )
                cache = self.cache_for_params(merged_params)
                return cache.cache_status(axis_indices=axis_indices, **axis_inputs)

            setattr(wrapper, "cache_status", cache_status)
            return wrapper

        return decorator


def _map_process_pool(
    func: Callable[..., Any], items: Iterable[Any], **kwargs: Any
) -> Iterable[Any]:
    max_workers = kwargs.get("max_workers")
    chunksize = kwargs.get("chunksize")
    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        if chunksize is None:
            return list(executor.map(func, items))
        return list(executor.map(func, items, chunksize=chunksize))


def _resolve_map_kwargs(
    map_fn_kwargs: Mapping[str, Any] | None, max_workers: int
) -> dict[str, Any]:
    resolved = dict(map_fn_kwargs or {})
    if max_workers > 1 and "max_workers" not in resolved:
        resolved["max_workers"] = max_workers
    return resolved


def _build_items_for_parallel(cache: ChunkCache) -> list[dict[str, Any]]:
    status = cache.cache_status()
    axis_values = status.get("axis_values", {})
    axis_order = status.get("axis_order", tuple(axis_values))
    axis_lists: list[list[Any]] = []
    for axis in axis_order:
        values = axis_values.get(axis, [])
        if isinstance(values, (list, tuple)):
            axis_lists.append(list(values))
        elif isinstance(values, Sequence):
            axis_lists.append(list(values))
        else:
            axis_lists.append([values])
    return [dict(zip(axis_order, values)) for values in itertools.product(*axis_lists)]


def _require_top_level_function(func: Callable[..., Any]) -> None:
    qualname = getattr(func, "__qualname__", "")
    if "<locals>" in qualname:
        raise ValueError(
            "Parallel execution requires a top-level function (module scope) so it can be "
            "pickled for ProcessPoolExecutor. Move the function to module scope or pass a "
            "custom map_fn that does not require pickling."
        )



from importlib.metadata import PackageNotFoundError, version

from .cache import ChunkCache
from .memo import ChunkMemo
from .identity import params_to_cache_id
from .runner_protocol import CacheStatus, RunnerContext
from .runners import (
    Diagnostics,
    run,
    run_parallel,
    run_streaming,
)

auto_load = ChunkMemo.auto_load


try:
    __version__ = version("chunk-memo")
except PackageNotFoundError:
    __version__ = "unknown"

__all__ = [
    "ChunkCache",
    "ChunkMemo",
    "Diagnostics",
    "__version__",
    "auto_load",
    "run",
    "run_parallel",
    "run_streaming",
    "RunnerContext",
    "CacheStatus",
    "params_to_cache_id",
]



import json
import os
import pickle
import tempfile
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Mapping


def _now_iso() -> str:
    return datetime.now(timezone.utc).isoformat()


def _apply_payload_timestamps(
    payload: dict[str, Any],
    existing: Mapping[str, Any] | None = None,
) -> None:
    now = _now_iso()
    created_at = None if existing is None else existing.get("created_at")
    if created_at is None:
        created_at = now
    payload["created_at"] = created_at
    payload["updated_at"] = now


def _atomic_write_pickle(path: Path, payload: dict[str, Any]) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    tmp_path: Path | None = None
    try:
        with tempfile.NamedTemporaryFile(
            mode="wb",
            delete=False,
            dir=path.parent,
            prefix=f".{path.name}.",
        ) as handle:
            tmp_path = Path(handle.name)
            pickle.dump(payload, handle, protocol=5)
        os.replace(tmp_path, path)
    finally:
        if tmp_path is not None and tmp_path.exists():
            tmp_path.unlink()


def _atomic_write_json(path: Path, payload: dict[str, Any]) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    tmp_path: Path | None = None
    try:
        with tempfile.NamedTemporaryFile(
            mode="w",
            delete=False,
            dir=path.parent,
            prefix=f".{path.name}.",
            encoding="utf-8",
        ) as handle:
            tmp_path = Path(handle.name)
            json.dump(payload, handle, ensure_ascii=True, sort_keys=True)
        os.replace(tmp_path, path)
    finally:
        if tmp_path is not None and tmp_path.exists():
            tmp_path.unlink()
